"""
Neuraxle's Automatic Machine Learning Classes
==================================================
All steps, and abstractions needed to build Automatic Machine Learning algorithms in Neuraxle.
..
    Copyright 2019, Neuraxio Inc.
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
        http://www.apache.org/licenses/LICENSE-2.0
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import copy
import glob
import hashlib
import json
import os
import time
import traceback
from abc import ABC, abstractmethod
from typing import Callable, List, Tuple
import numpy as np

from sklearn.metrics import r2_score

from neuraxle.base import MetaStep, BaseStep, ExecutionContext, ForceHandleOnlyMixin, \
    EvaluableStepMixin, _FittableStep
from neuraxle.data_container import DataContainer
from neuraxle.hyperparams.space import HyperparameterSamples, HyperparameterSpace

from neuraxle.metaopt.auto_ml import RandomSearchHyperparameterSelectionStrategy
from neuraxle.metaopt.observable import _Observable
from neuraxle.metaopt.trial import TRIAL_STATUS, Trial
from neuraxle.metaopt.random import BaseCrossValidationWrapper, BaseValidation
from neuraxle.steps.loop import StepClonerForEachDataInput
from neuraxle.steps.numpy import NumpyConcatenateOuterBatch

class AutoMLContainer:
    """
    Data object for auto ml.

    .. seealso::
        :class:`Trainer`,
        :class:`~neuraxle.metaopt.trial.Trial`,
        :class:`~neuraxle.metaopt.trial.Trials`,
        :class:`HyperparamsRepository`,
        :class:`InMemoryHyperparamsRepository`,
        :class:`HyperparamsJSONRepository`,
        :class:`BaseHyperparameterSelectionStrategy`,
        :class:`RandomSearchHyperparameterSelectionStrategy`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSamples`
    """

    def __init__(
            self,
            trials: 'Trials',
            hyperparameter_space: HyperparameterSpace,
            trial_number: int
    ):
        self.trials = trials
        self.hyperparameter_space = hyperparameter_space
        self.trial_number = trial_number


class HyperparamsRepository(_Observable[Trial], ABC):
    """
    Hyperparams repository that saves hyperparams, and scores for every AutoML trial.

    .. seealso::
        :class:`AutoMLSequentialWrapper`
        :class:`AutoMLAlgorithm`,
        :class:`BaseValidation`,
        :class:`RandomSearchBaseAutoMLStrategy`,
        :class:`HyperparameterSpace`,
        :class:`HyperparameterSamples`
    """

    @abstractmethod
    def create_new_trial(self, hyperparams: HyperparameterSamples):
        """
        Create new hyperperams trial.

        :param hyperparams: hyperparams
        :type hyperparams: HyperparameterSamples
        :return:
        """
        pass

    @abstractmethod
    def load_all_trials(self, status: 'TRIAL_STATUS') -> 'Trials':
        """
        Load all hyperparameter trials with their corresponding score.
        Sorted by creation date.

        :return: Trials (hyperparams, scores)
        """
        pass

    @abstractmethod
    def save_score_for_success_trial(self, hyperparams: HyperparameterSamples, score: float):
        """
        Save hyperparams, and score for a successful trial.

        :return: (hyperparams, scores)
        """
        pass

    @abstractmethod
    def save_failure_for_trial(self, hyperparams: HyperparameterSamples, exception: Exception):
        """
        Save hyperparams, and score for a failed trial.

        :return: (hyperparams, scores)
        """
        pass

    def new_experiment(self, hyperparams: HyperparameterSamples):
        pass

    def _get_trial_hash(self, hp_dict):
        current_hyperparameters_hash = hashlib.md5(str.encode(str(hp_dict))).hexdigest()
        return current_hyperparameters_hash


class InMemoryHyperparamsRepository(HyperparamsRepository):
    """
    In memory hyperparams repository that can print information about trials.
    Useful for debugging.

    Example usage :

    .. code-block:: python

        InMemoryHyperparamsRepository(
            print_new_trial=True,
            print_success_trial=True,
            print_exception=True
        )

    .. seealso::
        :class:`HyperparamsRepository`,
        :class:`HyperparameterSamples`,
        :class:`AutoMLSequentialWrapper`
    """

    def __init__(self, print_new_trial=True, print_success_trial=True, print_exception=True,
                 print_func: Callable = None):
        HyperparamsRepository.__init__(self)
        if print_func is None:
            print_func = print
        self.print_func = print_func

        self.trials = Trials()
        self.print_new_trial = print_new_trial
        self.print_success_trial = print_success_trial
        self.print_exception = print_exception

    def create_new_trial(self, hyperparams: HyperparameterSamples):
        if self.print_new_trial:
            self.print_func('new trial:\n{}'.format(json.dumps(hyperparams.to_nested_dict(), sort_keys=True, indent=4)))

    def load_all_trials(self, status: 'TRIAL_STATUS' = None) -> 'Trials':
        return self.trials.filter(status)

    def save_score_for_success_trial(self, hyperparams: HyperparameterSamples, score: float):
        self.trials.append(Trial(hyperparams, score, TRIAL_STATUS.SUCCESS))

        if self.print_success_trial:
            self.print_func('score: {}'.format(score))
            self.print_func(
                'hyperparams:\n{}'.format(json.dumps(hyperparams.to_nested_dict(), sort_keys=True, indent=4)))

    def save_failure_for_trial(self, hyperparams: HyperparameterSamples, exception: Exception):
        if self.print_exception:
            self.print_func(exception)
            traceback_str = str(traceback.format_tb(exception.__traceback__))
            self.print_func(traceback_str)


class HyperparamsJSONRepository(HyperparamsRepository):
    """
    Hyperparams repository that saves json files for every AutoML trial.

    .. seealso::
        :class:`HyperparamsRepository`,
        :class:`HyperparameterSamples`,
        :class:`AutoMLSequentialWrapper`
    """

    def __init__(self, folder):
        self.cache_folder = folder

    def create_new_trial(self, hyperparams: HyperparameterSamples):
        """
        Create new hyperperams trial json file.

        :param hyperparams: hyperparams
        :type hyperparams: HyperparameterSamples
        :return:
        """
        self._create_trial_json(hyperparams)

    def load_all_trials(self, status: 'TRIAL_STATUS' = None) -> 'Trials':
        """
        Load all hyperparameter trials with their corresponding score.
        Reads all the saved trial json files, sorted by creation date.

        :return: (hyperparams, scores)
        """
        trials = Trials()

        files = glob.glob(os.path.join(self.cache_folder, '*.json'))

        # sort by created date:
        def getmtimens(filename):
            return os.stat(filename).st_mtime_ns

        files.sort(key=getmtimens)

        for base_path in files:
            with open(base_path) as f:
                trial_json = json.load(f)

            if status is None or trial_json['status'] == status.value:
                trials.append(Trial.from_json(trial_json))

        return trials

    def save_score_for_success_trial(self, hyperparams: HyperparameterSamples, score: float):
        """
        Save hyperparams, and score for a successful trial.

        :return: (hyperparams, scores)
        """
        self._save_successful_trial_json(hyperparams, score)

    def save_failure_for_trial(self, hyperparams: HyperparameterSamples, exception: Exception):
        """
        Save hyperparams, and score for a failed trial.

        :return: (hyperparams, scores)
        """
        self._save_failed_trial_json(hyperparams, exception)

    def _create_trial_json(self, hyperparams):
        """
        Save new trial json file.

        :return: (hyperparams, scores)
        """
        hp_dict = hyperparams.to_flat_as_dict_primitive()
        current_hyperparameters_hash = self._get_trial_hash(hp_dict)

        with open(os.path.join(self._get_new_trial_json_path(current_hyperparameters_hash)), 'w+') as outfile:
            json.dump({
                'hyperparams': hp_dict,
                'score': None,
                'status': TRIAL_STATUS.PLANNED.value
            }, outfile)

    def _save_successful_trial_json(self, hyperparams, score):
        """
        Save trial json file.

        :return: (hyperparams, scores)
        """
        hp_dict = hyperparams.to_flat_as_dict_primitive()
        current_hyperparameters_hash = self._get_trial_hash(hp_dict)
        self._remove_new_trial_json(current_hyperparameters_hash)

        with open(os.path.join(self.cache_folder,
                               str(float(score)).replace('.', ',') + "_" + current_hyperparameters_hash) + '.json',
                  'w+') as outfile:
            json.dump({
                'hyperparams': hp_dict,
                'score': score,
                'status': TRIAL_STATUS.SUCCESS.value
            }, outfile)
        # Sleeping to have a valid time difference between files when reloading them to sort them by creation time:
        time.sleep(0.1)

    def _remove_new_trial_json(self, current_hyperparameters_hash):
        new_trial_json = self._get_new_trial_json_path(current_hyperparameters_hash)
        if os.path.exists(new_trial_json):
            os.remove(new_trial_json)

    def _get_new_trial_json_path(self, current_hyperparameters_hash):
        return os.path.join(self.cache_folder, "NEW_" + current_hyperparameters_hash) + '.json'

    def _save_failed_trial_json(self, hyperparams, exception):
        """
        Save trial json file.

        :return: (hyperparams, scores)
        """
        hp_dict = hyperparams.to_flat_as_dict_primitive()
        current_hyperparameters_hash = self._get_trial_hash(hp_dict)
        self._remove_new_trial_json(current_hyperparameters_hash)

        with open(os.path.join(self.cache_folder, 'FAILED_' + current_hyperparameters_hash) + '.json', 'w+') as outfile:
            json.dump({
                'hyperparams': hp_dict,
                'score': None,
                'status': TRIAL_STATUS.FAILED.value,
                'exception': str(exception)
            }, outfile)


class BaseHyperparameterOptimizer(ABC):
    @abstractmethod
    def find_next_best_hyperparams(self, auto_ml_container: 'AutoMLContainer') -> HyperparameterSamples:
        """
        Find the next best hyperparams using previous trials.

        :param auto_ml_container: trials data container
        :type auto_ml_container: neuraxle.metaopt.new_automl.Trials
        :return: next best hyperparams
        :rtype: HyperparameterSamples
        """
        raise NotImplementedError()


class AutoMLAlgorithm(ForceHandleOnlyMixin, MetaStep):
    """
    Pipeline step that executes Automatic Machine Learning strategy.
    It uses an hyperparameter optimizer of type :class:`BaseHyperparameterOptimizer` to find the next best hyperparams.
    It uses a validation technique of type :class:`BaseCrossValidationWrapper` to calculate the score.
    Please refer to :class:`AutoMLSequentialWrapper` for a usage example.

    .. seealso::
        :class:`BaseCrossValidationWrapper`,
        :class:`HyperparameterSamples`,
        :class:`BaseHyperparameterOptimizer`,
        :class:`BaseCrossValidationWrapper`,
        :class:`TrialsContainer`,
        :class:`HyperparameterSpace`,
        :class:`MetaStepMixin`,
        :class:`BaseStep`
    """

    def __init__(
            self,
            hyperparameter_optimizer: BaseHyperparameterOptimizer,
            higher_score_is_better=True,
            cache_folder_when_no_handle=None
    ):
        MetaStep.__init__(self, None)
        ForceHandleOnlyMixin.__init__(self, cache_folder_when_no_handle)
        self.higher_score_is_better = higher_score_is_better
        self.hyperparameter_optimizer = hyperparameter_optimizer

    def find_next_best_hyperparams(self, auto_ml_container: 'AutoMLContainer') -> HyperparameterSamples:
        """
        Find the next best hyperparams using previous trials.

        :param auto_ml_container: trials data container
        :type auto_ml_container: neuraxle.metaopt.new_automl.Trials
        :return: next best hyperparams
        :rtype: HyperparameterSamples
        """
        return self.hyperparameter_optimizer.find_next_best_hyperparams(auto_ml_container)

    def _fit_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> (
            'BaseStep', DataContainer):
        """
        Fit cross validation with wrapped step, and return the score.

        :param data_container: data container to fit transform
        :param context: execution context
        :return: self, step score
        :rtype: (AutoMLAlgorithm, float)
        """
        step: BaseCrossValidationWrapper = copy.deepcopy(self.wrapped)
        step = step.handle_fit(data_container, context)
        score = step.get_score()

        data_container.set_data_inputs(score)

        return self, data_container

    def get_best_hyperparams(self, trials: 'Trials') -> HyperparameterSamples:
        """
        Get the best hyperparams from all previous trials.

        :return: best hyperparams
        :rtype: HyperparameterSamples
        """
        return trials.get_best_hyperparams(higher_score_is_better=self.higher_score_is_better)

    def fit(self, data_inputs, expected_outputs=None):
        return self

    def transform(self, data_inputs, expected_outputs=None):
        return data_inputs


class Trial:
    def __init__(self, hyperparams: HyperparameterSamples, score: float, status: TRIAL_STATUS):
        self.hyperparams = HyperparameterSamples(hyperparams)
        self.score = score
        self.status = status

    def to_json(self) -> dict:
        return {
            'hyperparams': self.hyperparams.to_flat_as_dict_primitive(),
            'score': self.score,
            'status': self.status
        }

    @staticmethod
    def from_json(trial_json) -> 'Trial':
        return Trial(
            hyperparams=trial_json['hyperparams'],
            score=trial_json['score'],
            status=trial_json['status']
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = "Trial.from_json({})".format(str(self.to_json()))
        return s


class AutoMLSequentialWrapper(ForceHandleOnlyMixin, MetaStep):
    """
    A step to execute any Automatic Machine Learning Algorithms.

    Example usage :

    .. code-block:: python

        auto_ml: AutoMLSequentialWrapper = AutoMLSequentialWrapper(
            step=ForecastingPipeline(),
            auto_ml_algorithm=AutoMLAlgorithm(
                hyperparameter_optimizer=RandomSearchHyperparameterOptimizer(),
                validation_technique=KFoldCrossValidationWrapper(),
                higher_score_is_better=True
            ),
            hyperparams_repository=HyperparamsJSONRepository(),
            n_iters=100
        )
        auto_ml: AutoMLSequentialWrapper = auto_ml.fit(data_inputs, expected_outputs)
        best_model: ForecastingPipeline = auto_ml.get_best_model()


    .. seealso::
        :class:`AutoMLAlgorithm`,
        :class:`HyperparamsRepository`,
        :class:`RandomSearch`,
        :class:`BaseHyperparameterOptimizer`,
        :class:`RandomSearchHyperparameterOptimizer`,
        :class:`MetaStepMixin`,
        :class:`NonTransformableMixin`,
        :class:`BaseStep`
    """

    def __init__(
            self,
            wrapped: BaseStep,
            auto_ml_algorithm: AutoMLAlgorithm,
            hyperparams_repository: HyperparamsRepository = None,
            n_iters: int = 100,
            refit=True,
            cache_folder_when_no_handle=None
    ):
        if not isinstance(wrapped, EvaluableStepMixin):
            raise ValueError(
                'AutoML algorithm needs evaluable steps that implement the function get_score. Please use a validation technique, or implement EvaluableStepMixin.')

        MetaStep.__init__(self, auto_ml_algorithm.set_step(wrapped))
        ForceHandleOnlyMixin.__init__(self, cache_folder_when_no_handle)

        if hyperparams_repository is None:
            hyperparams_repository = InMemoryHyperparamsRepository()
        self.hyperparams_repository = hyperparams_repository
        self.n_iters = n_iters
        self.refit = refit

    def set_step(self, step: BaseStep) -> BaseStep:
        auto_ml_algorithm: AutoMLAlgorithm = self.get_step()
        auto_ml_algorithm.set_step(step)  # 2nd level wrapped set shortcut.
        return self

    def _fit_transform_data_container(self, data_container, context):
        new_self = self._fit_data_container(data_container, context)

        new_self.best_model.apply('disable_metrics')
        data_container = new_self._transform_data_container(data_container, context)
        new_self.best_model.apply('enable_metrics')

        return new_self, data_container

    def _fit_data_container(self, data_container: DataContainer, context: ExecutionContext):
        """
        Find the best hyperparams using the wrapped AutoML strategy.

        :param data_container: data container to fit on
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: fitted self
        :rtype: BaseStep
        """
        for i in range(self.n_iters):
            auto_ml_trial_data_container: AutoMLContainer = self._load_auto_ml_data(i)

            hyperparams = self.wrapped.find_next_best_hyperparams(auto_ml_trial_data_container)
            self.wrapped = self.wrapped.update_hyperparams(hyperparams)

            self.hyperparams_repository.create_new_trial(hyperparams)

            try:
                self.wrapped, data_container_with_score = self.wrapped.handle_fit_transform(data_container.copy(),
                                                                                            context)
                score = data_container_with_score.data_inputs

                self.hyperparams_repository.save_score_for_success_trial(hyperparams, score)
            except Exception as error:
                track = traceback.format_exc()
                print(track)
                self.hyperparams_repository.save_failure_for_trial(hyperparams, error)

        if self.refit:
            best_model = self._load_virgin_best_model()
            best_model = best_model.handle_fit(data_container.copy(), context)
            self.best_model = best_model

        return self

    def _transform_data_container(self, data_container, context):
        if self.best_model is None:
            raise Exception('Cannot transform AutoMLSequentialWrapper before fit')
        return self.best_model.handle_transform(data_container, context)

    def transform(self, data_inputs):
        if self.best_model is None:
            raise Exception('Cannot transform AutoMLSequentialWrapper before fit')

        return self.best_model.transform(data_inputs)

    def _load_auto_ml_data(self, trial_number: int) -> AutoMLContainer:
        """
        Load data for all trials.

        :param trial_number: trial number
        :type trial_number: int
        :return: auto ml data container
        :rtype: neuraxle.metaopt.new_automl.Trials
        """
        trials = self.hyperparams_repository.load_all_trials(TRIAL_STATUS.SUCCESS)

        return AutoMLContainer(
            trial_number=trial_number,
            trials=trials,
            hyperparameter_space=self.wrapped.get_hyperparams_space()
        )

    def _load_virgin_best_model(self) -> BaseStep:
        """
        Get the best model from all of the previous trials.
        :return: best model step
        :rtype: BaseStep
        """
        trials: Trials = self.hyperparams_repository.load_all_trials(TRIAL_STATUS.SUCCESS)
        auto_ml_algorithm: AutoMLAlgorithm = self.wrapped
        best_hyperparams = auto_ml_algorithm.get_best_hyperparams(trials)
        auto_ml_algorithm = auto_ml_algorithm.update_hyperparams(best_hyperparams)

        validation_technique = auto_ml_algorithm.get_step()
        best_model = validation_technique.get_step()

        return copy.deepcopy(best_model)

    def get_best_model(self) -> BaseStep:
        return self.best_model


class RandomSearch(AutoMLSequentialWrapper):
    """
    Random Search Automatic Machine Learning Algorithm that randomly samples the space of random variables.

    Example usage :

    .. code-block:: python

        random_search = RandomSearch(
            ForecastingPipeline(),
            hyperparams_repository=HyperparamsJSONRepository(),
            n_iters=100
        )


    .. seealso::
        :class:`AutoMLSequentialWrapper`,
        :class:`AutoMLAlgorithm`,
        :class:`HyperparamsRepository`,
        :class:`InMemoryHyperparamsRepository`,
        :class:`HyperparamsJSONRepository`,
        :class:`RandomSearch`,
        :class:`BaseHyperparameterOptimizer`,
        :class:`RandomSearchHyperparameterOptimizer`
    """

    def __init__(
            self,
            wrapped: BaseStep = None,
            hyperparams_repository: HyperparamsRepository = None,
            higher_score_is_better=True,
            n_iter: int = 10,
            cache_folder_when_no_handle=None
    ):
        AutoMLSequentialWrapper.__init__(
            self,
            wrapped=wrapped,
            auto_ml_algorithm=AutoMLAlgorithm(
                hyperparameter_optimizer=RandomSearchHyperparameterSelectionStrategy(),
                higher_score_is_better=higher_score_is_better,
                cache_folder_when_no_handle=cache_folder_when_no_handle
            ),
            hyperparams_repository=hyperparams_repository,
            n_iters=n_iter
        )


class Trial:
    def __init__(self, hyperparams: HyperparameterSamples, score: float, status: TRIAL_STATUS):
        self.hyperparams = HyperparameterSamples(hyperparams)
        self.score = score
        self.status = status

    def to_json(self) -> dict:
        return {
            'hyperparams': self.hyperparams.to_flat_as_dict_primitive(),
            'score': self.score,
            'status': self.status
        }

    @staticmethod
    def from_json(trial_json) -> 'Trial':
        return Trial(
            hyperparams=trial_json['hyperparams'],
            score=trial_json['score'],
            status=trial_json['status']
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = "Trial.from_json({})".format(str(self.to_json()))
        return s


class Trials:
    """
    Data object containing auto ml trials.
    .. seealso::
        :class:`AutoMLSequentialWrapper`,
        :class:`RandomSearch`,
        :class:`HyperparamsRepository`,
        :class:`MetaStepMixin`,
        :class:`BaseStep`
    """

    def __init__(
            self,
            trials: List[Trial] = None
    ):
        if trials is None:
            trials = []
        self.trials: List[Trial] = trials

    def get_best_hyperparams(self, higher_score_is_better: bool) -> HyperparameterSamples:
        best_score = None
        best_hyperparams = None

        for trial in self.trials:
            if best_score is None or higher_score_is_better == (trial.score > best_score):
                best_score = trial.score
                best_hyperparams = trial.hyperparams

        return best_hyperparams

    def append(self, trial: Trial):
        self.trials.append(trial)

    def filter(self, status: TRIAL_STATUS) -> 'Trials':
        trials = Trials()
        for trial in self.trials:
            if trial.status == status:
                trials.append(trial)

        return trials

    def __getitem__(self, item):
        return self.trials[item]

    def __len__(self):
        return len(self.trials)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        s = "Trials({})".format(str([str(t) for t in self.trials]))
        return s


class BaseCrossValidationWrapper(EvaluableStepMixin, ForceHandleOnlyMixin, BaseValidation, ABC):
    # TODO: change default argument of scoring_function...
    def __init__(self, wrapped=None, scoring_function=r2_score, joiner=NumpyConcatenateOuterBatch(),
                 cache_folder_when_no_handle=None,
                 split_data_container_during_fit=True, predict_after_fit=True):
        BaseValidation.__init__(self, wrapped=wrapped, scoring_function=scoring_function)
        ForceHandleOnlyMixin.__init__(self, cache_folder=cache_folder_when_no_handle)
        EvaluableStepMixin.__init__(self)

        self.split_data_container_during_fit = split_data_container_during_fit
        self.predict_after_fit = predict_after_fit
        self.joiner = joiner

    def train(self, train_data_container: DataContainer, context: ExecutionContext):
        step = StepClonerForEachDataInput(self.wrapped)
        step = step.handle_fit(train_data_container, context)

        return step

    def _fit_data_container(self, data_container: DataContainer, context: ExecutionContext) -> BaseStep:
        assert self.wrapped is not None

        if self.split_data_container_during_fit:
            train_data_container, validation_data_container = self.split_data_container(data_container)
        else:
            train_data_container = data_container

        step = StepClonerForEachDataInput(self.wrapped)
        step = step.handle_fit(train_data_container, context)

        if self.predict_after_fit:
            results = step.handle_predict(validation_data_container, context)
            self.calculate_score(results)

        return self

    def calculate_score(self, results):
        self.scores = [self.scoring_function(a, b) for a, b in zip(results.data_inputs, results.expected_outputs)]
        self.scores_mean = np.mean(self.scores)
        self.scores_std = np.std(self.scores)

    def split_data_container(self, data_container: DataContainer) -> Tuple[DataContainer, DataContainer]:
        train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs = self.split(
            data_container.data_inputs,
            data_container.expected_outputs
        )

        train_data_container = DataContainer(data_inputs=train_data_inputs, expected_outputs=train_expected_outputs)
        validation_data_container = DataContainer(data_inputs=validation_data_inputs,
                                                  expected_outputs=validation_expected_outputs)

        return train_data_container, validation_data_container

    def get_score(self):
        return self.scores_mean

    def get_scores_std(self):
        return self.scores_std

    @abstractmethod
    def split(self, data_inputs, expected_outputs):
        raise NotImplementedError("TODO")


class KFoldCrossValidationWrapper(BaseCrossValidationWrapper):
    def __init__(
            self,
            scoring_function=r2_score,
            k_fold=3,
            joiner=NumpyConcatenateOuterBatch(),
            cache_folder_when_no_handle=None,
            split_data_container_during_fit=True,
            predict_after_fit=True
    ):
        self.k_fold = k_fold
        BaseCrossValidationWrapper.__init__(
            self,
            scoring_function=scoring_function,
            joiner=joiner,
            cache_folder_when_no_handle=cache_folder_when_no_handle,
            split_data_container_during_fit=split_data_container_during_fit,
            predict_after_fit=predict_after_fit
        )

    def split(self, data_inputs, expected_outputs):
        validation_data_inputs, validation_expected_outputs = self.validation_split(data_inputs, expected_outputs)
        train_data_inputs, train_expected_outputs = self.train_split(data_inputs, expected_outputs)

        return train_data_inputs, train_expected_outputs, validation_data_inputs, validation_expected_outputs

    def train_split(self, data_inputs, expected_outputs) -> (List, List):
        train_data_inputs = []
        train_expected_outputs = []
        data_inputs = np.array(data_inputs)
        expected_outputs = np.array(expected_outputs)

        for i in range(len(data_inputs)):
            before_di = data_inputs[:i]
            after_di = data_inputs[i + 1:]
            inputs = (before_di, after_di)

            before_eo = expected_outputs[:i]
            after_eo = expected_outputs[i + 1:]
            outputs = (before_eo, after_eo)

            inputs = self.joiner.transform(inputs)
            outputs = self.joiner.transform(outputs)

            train_data_inputs.append(inputs)
            train_expected_outputs.append(outputs)

        return train_data_inputs, train_expected_outputs

    def validation_split(self, data_inputs, expected_outputs=None) -> (List, List):
        splitted_data_inputs = self._split(data_inputs)
        if expected_outputs is not None:
            splitted_expected_outputs = self._split(expected_outputs)
            return splitted_data_inputs, splitted_expected_outputs

        return splitted_data_inputs, [None] * len(splitted_data_inputs)

    def _split(self, data_inputs):
        splitted_data_inputs = []
        step = len(data_inputs) / float(self.k_fold)
        for i in range(self.k_fold):
            a = int(step * i)
            b = int(step * (i + 1))
            if b > len(data_inputs):
                b = len(data_inputs)

            slice = data_inputs[a:b]
            splitted_data_inputs.append(slice)
        return splitted_data_inputs





"""
Neuraxle's Pipeline Classes
====================================
This is the core of Neuraxle's pipelines. You can chain steps to call them one after an other.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import shutil
import warnings
from abc import ABC, abstractmethod
from copy import copy
from typing import Any, Tuple, List

from neuraxle.base import BaseStep, TruncableSteps, NamedTupleList, ResumableStepMixin, ExecutionContext, ExecutionMode, \
    NonTransformableMixin, MetaStep, _FittableStep, HandleOnlyMixin, ForceHandleOnlyMixin, _CustomHandlerMethods, \
    ForceHandleMixin, Identity
from neuraxle.checkpoints import Checkpoint
from neuraxle.data_container import DataContainer, ListDataContainer

DEFAULT_CACHE_FOLDER = 'cache'


class BasePipeline(TruncableSteps, ABC):
    def __init__(self, steps: NamedTupleList):
        TruncableSteps.__init__(self, steps_as_tuple=steps)

    @abstractmethod
    def fit(self, data_inputs, expected_outputs=None) -> 'BasePipeline':
        raise NotImplementedError()

    @abstractmethod
    def transform(self, data_inputs):
        raise NotImplementedError()

    @abstractmethod
    def fit_transform(self, data_inputs, expected_outputs=None) -> ('BasePipeline', Any):
        raise NotImplementedError()


class Pipeline(BasePipeline):
    """
    Fits and transform steps
    """

    def __init__(self, steps: NamedTupleList, cache_folder=DEFAULT_CACHE_FOLDER):
        BasePipeline.__init__(self, steps=steps)
        self.cache_folder = cache_folder

    def transform(self, data_inputs: Any):
        """
        After loading the last checkpoint, transform each pipeline steps

        :param data_inputs: the data input to transform
        :return: transformed data inputs
        """
        data_container = self.transform_data_container(DataContainer(data_inputs=data_inputs, current_ids=None))
        return data_container.data_inputs

    def transform_data_container(self, data_container: DataContainer):
        data_container = self.handle_transform(
            data_container,
            ExecutionContext(root=self.cache_folder, execution_mode=ExecutionMode.TRANSFORM)
        )
        return data_container

    def fit_transform(self, data_inputs, expected_outputs=None) -> ('Pipeline', Any):
        """
        After loading the last checkpoint, fit transform each pipeline steps

        :param data_inputs: the data input to fit on
        :param expected_outputs: the expected data output to fit on
        :return: the pipeline itself
        """
        new_self, data_container = self.fit_transform_data_container(DataContainer(data_inputs=data_inputs, current_ids=None, expected_outputs=expected_outputs))
        return new_self, data_container.data_inputs

    def fit_transform_data_container(self, data_container) -> Tuple['Pipeline', DataContainer]:
        context = ExecutionContext(root=self.cache_folder, execution_mode=ExecutionMode.FIT_TRANSFORM)
        new_self, data_container = self.handle_fit_transform(data_container, context)

        return new_self, data_container

    def fit(self, data_inputs, expected_outputs=None) -> 'Pipeline':
        """
        After loading the last checkpoint, fit each pipeline steps

        :param data_inputs: the data input to fit on
        :param expected_outputs: the expected data output to fit on
        :return: the pipeline itself
        """
        return self.fit_data_container(DataContainer(data_inputs=data_inputs, current_ids=None, expected_outputs=expected_outputs))

    def fit_data_container(self, data_container) -> 'Pipeline':
        return self.handle_fit(data_container, ExecutionContext(self.cache_folder, ExecutionMode.FIT))

    def inverse_transform(self, processed_outputs) -> Any:
        """
        After transforming all data inputs, and obtaining a prediction, we can inverse transform the processed outputs

        :param processed_outputs: the forward transformed data input
        :return: backward transformed processed outputs
        """
        data_container = DataContainer(data_inputs=processed_outputs)
        context = ExecutionContext(root=self.cache_folder, execution_mode=ExecutionMode.INVERSE_TRANSFORM)

        for step_name, step in list(reversed(self.items())):
            data_container = step.handle_inverse_transform(data_container, context)

        return data_container.data_inputs

    def _inverse_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        After transforming all data inputs, and obtaining a prediction, we can inverse transform the processed outputs
        """
        for step_name, step in list(reversed(self.items())):
            data_container = step.handle_inverse_transform(data_container, context)
        return data_container

    def _fit_data_container(self, data_container: DataContainer, context: ExecutionContext) -> 'Pipeline':
        """
        After loading the last checkpoint, fit transform each pipeline steps,
        but only fit the last pipeline step.

        :param data_container: the data container to fit transform on
        :param context: execution context
        :return: tuple(pipeline, data_container)
        """
        steps_left_to_do, data_container = self._load_checkpoint(data_container, context)
        self.setup(context=context)

        index_last_step = len(steps_left_to_do) - 1

        new_steps_as_tuple: NamedTupleList = []

        for index, (step_name, step) in enumerate(steps_left_to_do):
            step.setup(context=context)

            if index != index_last_step:
                step, data_container = step.handle_fit_transform(data_container, context)
            else:
                step = step.handle_fit(data_container, context)

            new_steps_as_tuple.append((step_name, step))

        self.steps_as_tuple = self.steps_as_tuple[
                              :len(self.steps_as_tuple) - len(steps_left_to_do)] + new_steps_as_tuple

        return self

    def _fit_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> ('Pipeline', DataContainer):
        """
        After loading the last checkpoint, fit transform each pipeline steps

        :param data_container: the data container to fit transform on
        :param context: execution context
        :return: tuple(pipeline, data_container)
        """
        steps_left_to_do, data_container = self._load_checkpoint(data_container, context)
        self.setup(context=context)

        new_steps_as_tuple: NamedTupleList = []

        for step_name, step in steps_left_to_do:
            step.setup(context=context)
            step, data_container = step.handle_fit_transform(data_container, context)
            new_steps_as_tuple.append((step_name, step))

        self.steps_as_tuple = self.steps_as_tuple[:len(self.steps_as_tuple) - len(steps_left_to_do)] + \
                              new_steps_as_tuple

        return self, data_container

    def _transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        After loading the last checkpoint, transform each pipeline steps

        :param data_container: the data container to transform
        :return: transformed data container
        """
        steps_left_to_do, data_container = self._load_checkpoint(data_container, context)

        for step_name, step in steps_left_to_do:
            data_container = step.handle_transform(data_container, context)

        return data_container

    def _load_checkpoint(self, data_container: DataContainer, context: ExecutionContext) -> Tuple[NamedTupleList, DataContainer]:
        """
        Try loading a pipeline cache with the passed data container.
        If pipeline cache loading succeeds, find steps left to do,
        and load the latest data container.

        :param data_container: the data container to resume
        :param context: the execution context to resume
        :return: tuple(steps left to do, last checkpoint data container)
        """
        return self.steps_as_tuple, data_container

    def flush_all_cache(self):
        shutil.rmtree(self.cache_folder)


class ResumablePipeline(ResumableStepMixin, Pipeline):
    """
    Fits and transform steps after latest checkpoint
    """

    def _load_checkpoint(
            self,
            data_container: DataContainer,
            context: ExecutionContext
    ) -> Tuple[NamedTupleList, DataContainer]:
        """
        Try loading a pipeline cache with the passed data container.
        If pipeline cache loading succeeds, find steps left to do,
        and load the latest data container.

        :param data_container: the data container to resume
        :param context: the execution context to resume
        :return: tuple(steps left to do, last checkpoint data container)
        """
        new_starting_step_index, starting_step_data_container = \
            self._get_starting_step_info(data_container, context)

        loading_context = context.copy()
        loading_context.pop()
        loaded_pipeline = self.load(loading_context)

        if not self.are_steps_before_index_the_same(loaded_pipeline, new_starting_step_index):
            return self.steps_as_tuple, data_container

        self._assign_loaded_pipeline_into_self(loaded_pipeline)

        step = self[new_starting_step_index]
        if isinstance(step, Checkpoint) or (isinstance(step, MetaStep) and isinstance(step.wrapped, Checkpoint)):
            starting_step_data_container = step.resume(starting_step_data_container, context)

        return self[new_starting_step_index:], starting_step_data_container

    def _assign_loaded_pipeline_into_self(self, loaded_self):
        self.steps_as_tuple = loaded_self.steps_as_tuple
        self._refresh_steps()
        self.hyperparams = loaded_self.hyperparams
        self.hyperparams_space = loaded_self.hyperparams_space

    def _get_starting_step_info(self, data_container: DataContainer, context: ExecutionContext) -> Tuple[
        int, DataContainer]:
        """
        Find the index of the latest step that can be resumed

        :param data_container: the data container to resume
        :return: index of the latest resumable step, data container at starting step
        """
        starting_step_data_container = copy(data_container)
        starting_step_context = copy(context)
        current_data_container = copy(data_container)
        index_latest_checkpoint = 0

        for index, (step_name, step) in enumerate(self.items()):
            if hasattr(step, 'should_resume') and step.should_resume(current_data_container.copy(), starting_step_context):
                index_latest_checkpoint = index
                starting_step_data_container = copy(current_data_container)

            current_data_container = step.hash_data_container(current_data_container)

        return index_latest_checkpoint, starting_step_data_container

    def should_resume(self, data_container: DataContainer, context: ExecutionContext) -> bool:
        """
        Return True if the pipeline has a saved checkpoint that it can resume from

        :param context: execution context
        :param data_container: the data container to resume
        :return: bool
        """
        context = context.push(self)
        for index, (step_name, step) in enumerate(reversed(self.items())):
            if hasattr(step, 'should_resume') and step.should_resume(data_container, context):
                return True

        return False


class MiniBatchSequentialPipeline(_CustomHandlerMethods, ForceHandleMixin, Pipeline):
    """
    Mini Batch Sequential Pipeline class to create a pipeline processing data inputs in batch.

    Provide a default batch size :

    .. code-block:: python

        sub_pipelines = [SomeStep()]
        pipeline = MiniBatchSequentialPipeline(sub_pipelines, batch_size=32)


    Or manually add a :class`Barrier` step to the mini batch sequential pipeline :

    .. code-block:: python

        sub_pipelines = [SomeStep(), Joiner(32)]
        pipeline = MiniBatchSequentialPipeline(sub_pipelines)


    .. seealso::
        :class:`Pipeline`,
        :class:`Barrier`,
        :class:`Joiner`,
        :class:`~neuraxle.data_container.DataContainer`,
        :class:`~neuraxle.base.ExecutionContext`
    """
    def __init__(self, steps: NamedTupleList, batch_size=None, cache_folder=None):
        Pipeline.__init__(self, steps=steps, cache_folder=cache_folder)
        ForceHandleMixin.__init__(self)
        self.__validate_barriers_batch_size(batch_size)
        self.__patch_missing_barrier(batch_size)
        self.__patch_barriers_batch_size(batch_size)

    def __validate_barriers_batch_size(self, batch_size):
        if batch_size is not None:
            return

        for _, step in self:
            if isinstance(step, Barrier):
                if step.batch_size is None:
                    raise Exception(
                        'Invalid Joiner batch size {}[{}]. Please provide a default batch size to MiniBatchSequentialPipeline, or add a batch size to {}[{}].'.format(
                            self.name, step.name, self.name, step.name))

    def __patch_barriers_batch_size(self, batch_size):
        if batch_size is None:
            return

        for _, step in self:
            if isinstance(step, Barrier):
                if step.batch_size is not None:
                    warnings.warn(
                        'Replacing {}[{}].batch_size by {}.batch_size.'.format(self.name, step.name, self.name))
                step.batch_size = batch_size

    def __patch_missing_barrier(self, batch_size):
        has_barrier = False

        for _, step in self:
            if isinstance(step, Barrier):
                has_barrier = True

        if not has_barrier:
            self.steps_as_tuple.append(('Joiner', Joiner(batch_size)))

        self._refresh_steps()

    def transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        Transform all sub pipelines splitted by the Barrier steps.
        :param data_container: data container to transform.
        :param context: execution context
        :return: data container
        """
        sub_pipelines = self._create_sub_pipelines()

        for sub_pipeline in sub_pipelines:
            barrier = sub_pipeline[-1]
            data_container = barrier.join_transform(
                step=sub_pipeline,
                data_container=data_container,
                context=context
            )

            data_container = self.hash_data_container(data_container)

        return data_container

    def fit_data_container(self, data_container: DataContainer, context: ExecutionContext) -> BaseStep:
        """
        Fit all sub pipelines splitted by the Barrier steps.
        :param data_container: data container to transform.
        :param context: execution context
        :return: data container
        """
        sub_pipelines = self._create_sub_pipelines()
        index_start = 0

        for sub_pipeline in sub_pipelines:
            sub_pipeline.setup(context=context)

            barrier = sub_pipeline[-1]
            sub_pipeline, data_container = barrier.join_fit_transform(
                step=sub_pipeline,
                data_container=data_container,
                context=context
            )
            current_ids = self.hash(data_container)
            data_container.set_current_ids(current_ids)

            new_self = self[:index_start] + sub_pipeline
            if index_start + len(sub_pipeline) < len(self):
                new_self += self[index_start + len(sub_pipeline):]

            self.steps_as_tuple = new_self.steps_as_tuple
            index_start += len(sub_pipeline)

        return self

    def fit_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> Tuple[BaseStep, DataContainer]:
        """
        Transform all sub pipelines splitted by the Barrier steps.
        :param data_container: data container to transform.
        :param context: execution context
        :return: data container
        """
        sub_pipelines = self._create_sub_pipelines()
        index_start = 0

        for sub_pipeline in sub_pipelines:
            sub_pipeline.setup(context=context)

            barrier = sub_pipeline[-1]
            sub_pipeline, data_container = barrier.join_fit_transform(
                step=sub_pipeline,
                data_container=data_container,
                context=context
            )

            data_container = self.hash_data_container(data_container)

            new_self = self[:index_start] + sub_pipeline
            if index_start + len(sub_pipeline) < len(self):
                new_self += self[index_start + len(sub_pipeline):]

            self.steps_as_tuple = new_self.steps_as_tuple
            index_start += len(sub_pipeline)

        return self, data_container

    def _create_sub_pipelines(self) -> List['MiniBatchSequentialPipeline']:
        """
        Create sub pipelines by splitting the steps by the join type name.
        :return: list of sub pipelines
        """
        sub_pipelines: List[MiniBatchSequentialPipeline] = self.split(Barrier)
        for sub_pipeline in sub_pipelines:
            if not sub_pipeline.ends_with(Barrier):
                raise Exception('At least one Barrier step needs to be at the end of a streaming pipeline.')

        return sub_pipelines


class Barrier(Identity, ABC):
    """
    A Barrier step to be used in a minibatch sequential pipeline. It forces all the
    data inputs to get to the barrier in a sub pipeline before going through to the next sub-pipeline.

    .. code-block:: python

        p = MiniBatchSequentialPipeline([
            SomeStep(),
            SomeStep(),
            Barrier(), # must be a concrete Barrier ex: Joiner()
            SomeStep(),
            SomeStep(),
            Barrier(), # must be a concrete Barrier ex: Joiner()
        ], batch_size=10)


    .. seealso::
        :class:`~neuraxle.base.NonTransformableMixin`,
        :class:`~neuraxle.base.BaseStep`
    """

    @abstractmethod
    def join_transform(self, step: TruncableSteps, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        Execute the given pipeline :func:`~neuraxle.pipeline.Pipeline.transform` with the given data container, and execution context.
        :param step: truncable steps to execute
        :type step: TruncableSteps
        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: transformed data container
        :rtype: DataContainer
        """
        raise NotImplementedError()

    @abstractmethod
    def join_fit_transform(self, step: Pipeline, data_container: DataContainer, context: ExecutionContext) -> Tuple['Any', DataContainer]:
        """
        Execute the given pipeline :func:`~neuraxle.pipeline.Pipeline.fit_transform` with the given data container, and execution context.
        :param step: truncable steps to execute
        :param data_container: data container
        :param context: execution context
        :return: (fitted step, transformed data container)
        """
        raise NotImplementedError()


class Joiner(Barrier):
    """
    A Special Barrier step that joins the transformed mini batches together with list.extend method.
    """

    def __init__(self, batch_size):
        Barrier.__init__(self)
        self.batch_size = batch_size

    def join_transform(self, step: Pipeline, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        Concatenate the pipeline transform output of each batch of self.batch_size together.
        :param step: pipeline to transform on
        :type step: Pipeline
        :param data_container: data container to transform
        :type data_container: DataContainer
        :param context: execution context
        :return: transformed data container
        :rtype: DataContainer
        """
        context = context.push(step)

        data_container_batches = data_container.convolved_1d(
            stride=self.batch_size,
            kernel_size=self.batch_size
        )

        output_data_container = ListDataContainer.empty()
        for data_container_batch in data_container_batches:
            output_data_container.concat(
                step._transform_data_container(data_container_batch, context)
            )

        return output_data_container

    def join_fit_transform(self, step: Pipeline, data_container: DataContainer, context: ExecutionContext) -> Tuple['Any', DataContainer]:
        """
        Concatenate the pipeline fit transform output of each batch of self.batch_size together.
        :param step: pipeline to fit transform on
        :type step: Pipeline
        :param data_container: data container to fit transform on
        :type data_container: DataContainer
        :param context: execution context
        :return: fitted self, transformed data inputs
        :rtype: Tuple[Any, DataContainer]
        """
        context = context.push(step)

        data_container_batches = data_container.convolved_1d(
            stride=self.batch_size,
            kernel_size=self.batch_size
        )

        output_data_container = ListDataContainer.empty()
        for data_container_batch in data_container_batches:
            step, data_container_batch = step._fit_transform_data_container(data_container_batch, context)
            output_data_container.concat(
                data_container_batch
            )

        return step, output_data_container




import math
from abc import abstractmethod, ABC
from typing import Optional, List, Any

import numpy as np
from scipy.integrate import quad
from scipy.special import factorial
from scipy.stats import rv_continuous, norm, rv_discrete, rv_histogram, truncnorm, randint

from neuraxle.hyperparams.distributions import HyperparameterDistribution, WrappedHyperparameterDistributions, \
    DiscreteHyperparameterDistribution, ContinuousHyperparameterDistrbution


def scipy_method(func):
    def wrapper(*args, **kwargs):
        self_in_args = args[0]
        self_in_args._override_scipy_methods()
        return func(*args, **kwargs)

    return wrapper


class ScipyDistributionWrapper(ABC):
    """
    Base class for a distribution that wraps a scipy distribution.

    Usage example:

    .. code-block:: python

        distribution = ScipyDistributionWrapper(
            scipy_distribution=rv_histogram(histogram=histogram),
            null_default_value=null_default_value
        )

    .. seealso::
        :class:`HyperparameterDistribution`
    """

    def __init__(self, scipy_distribution, **scipy_distribution_arguments):
        self.scipy_distribution = scipy_distribution
        self.scipy_distribution_arguments = scipy_distribution_arguments

    def _override_scipy_methods(self):
        return

    @scipy_method
    def rvs(self, *args, **kwargs) -> float:
        if kwargs is None:
            kwargs = dict()
        kwargs = dict(kwargs)
        kwargs.update(**self.scipy_distribution_arguments)
        return self.scipy_distribution.rvs(*args, **kwargs)

    @scipy_method
    def pdf(self, x, *args, **kwargs) -> float:
        if hasattr(self.scipy_distribution, 'pdf'):
            return self.scipy_distribution.pdf(x, *args, **kwargs, **self.scipy_distribution_arguments)
        else:
            return self.scipy_distribution.pmf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def cdf(self, x, *args, **kwargs) -> float:
        """
        Cumulative distribution function of the given x.

        :param x:
        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.cdf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def entropy(self, *args, **kwargs):
        """
        Differential entropy of the RV.
        
        :return: 
        """
        return self.scipy_distribution.entropy(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def expect(self, *args, **kwargs):
        """
        Calculate expected value of a function with respect to the distribution by numerical integration.

        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.expect(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def fit(self, data, *args, **kwargs):
        """
        Return MLEs for shape( if applicable), location, and scale parameters from data.

        :param data:
        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.fit(data, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def fit_loc_scale(self, data, *args):
        """
        Estimate loc and scale parameters from data using 1 st and 2 nd moments. 
        
        :param data: 
        :return: 
        """
        return self.scipy_distribution.fit_loc_scale(data, *args, **self.scipy_distribution_arguments)

    @scipy_method
    def freeze(self, *args, **kwargs):
        """
        Freeze the distribution for the given arguments.

        :return:
        """
        return self.scipy_distribution.freeze(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def interval(self, alpha, *args, **kwargs):
        """
        Confidence interval with equal areas around the median.

        :param alpha:
        :return:
        """
        return self.scipy_distribution.interval(alpha, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def isf(self, q, *args, **kwargs):
        """
        Inverse survival function(inverse of sf) at q of the given RV.

        :param q:
        :return:
        """
        return self.scipy_distribution.isf(q, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def logcdf(self, x, *args, **kwargs):
        """
        Log of the cumulative distribution function at x of the given RV.

        :param x:
        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.logcdf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def logpdf(self, x, *args, **kwargs):
        """
        Log of the probability density function at x of the given RV.

        :param x:
        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.logpdf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def logsf(self, x, *args, **kwargs):
        """
        Log of the survival function of the given RV.

        :param x:
        :return:
        """
        return self.scipy_distribution.logsf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def min(self):
        return self.scipy_distribution.a

    @scipy_method
    def max(self):
        return self.scipy_distribution.b

    @scipy_method
    def mean(self, *args, **kwargs):
        """
        Mean of the distribution.

        :return:
        """
        return self.scipy_distribution.mean(*args, **kwargs)

    @scipy_method
    def median(self, *args, **kwargs):
        """
        Median of the distribution.

        :return:
        """
        return self.scipy_distribution.median(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def moment(self, n, *args, **kwargs):
        """
        n-th order non-central moment of distribution.

        :param n:
        :return:
        """
        return self.scipy_distribution.moment(n, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def nnlf(self, theta, x):
        """
        Return negative loglikelihood function.

        :param theta:
        :param x:
        :return:
        """
        return self.scipy_distribution.nnlf(theta, x)

    @scipy_method
    def ppf(self, q, *args, **kwargs):
        """
        Percent point function(inverse of cdf) at q of the given RV.

        :param q:
        :return:
        """
        return self.scipy_distribution.ppf(q, *args, **kwargs)

    @scipy_method
    def sf(self, x, *args, **kwargs):
        """
        Survival function(1 - cdf) at x of the given RV.

        :param x:
        :return:
        """
        return self.scipy_distribution.sf(x, *args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def stats(self, *args, **kwargs):
        """
        Some statistics of the given RV.

        :return:
        """
        return self.scipy_distribution.stats(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def std(self, *args, **kwargs):
        """
        Standard deviation of the distribution.

        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.std(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def support(self, *args, **kwargs):
        """
        Return the support of the distribution.

        :param args:
        :param kwargs:
        :return:
        """
        return self.scipy_distribution.support(*args, **kwargs, **self.scipy_distribution_arguments)

    @scipy_method
    def var(self, *args, **kwargs):
        """
        Variance of the distribution.

        :return:
        """
        return self.scipy_distribution.var(*args, **kwargs, **self.scipy_distribution_arguments)

    def to_sk_learn(self):
        return self.scipy_distribution


class ScipyContinuousDistributionWrapper(ScipyDistributionWrapper, HyperparameterDistribution):
    def __init__(self, scipy_distribution, null_default_value=None, **kwargs):
        ScipyDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_distribution,
            **kwargs
        )
        ContinuousHyperparameterDistrbution.__init__(self, null_default_value=null_default_value)


class ScipyDiscreteDistributionWrapper(ScipyDistributionWrapper, DiscreteHyperparameterDistribution):
    def __init__(self, scipy_distribution, null_default_value=None, **kwargs):
        ScipyDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_distribution,
            **kwargs
        )
        DiscreteHyperparameterDistribution.__init__(self, null_default_value=null_default_value)


class BaseCustomDiscreteScipyDistribution(ScipyDiscreteDistributionWrapper):
    def __init__(self, name, min_included, max_included, null_default_value, **kwargs):
        scipy_dist_obj = rv_discrete(
            name=name,
            a=min_included,
            b=max_included,
            **kwargs
        )
        scipy_dist_obj._pmf = self._pmf
        ScipyDiscreteDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_dist_obj,
            null_default_value=null_default_value
        )

    def _override_scipy_methods(self):
        self.scipy_distribution._pmf = self._pmf

    @abstractmethod
    def _pmf(self, x, *args):
        pass


class Distribution(rv_continuous):
    def _pdf(self, x):
        return 0.0


class BaseCustomContinuousScipyDistribution(ScipyContinuousDistributionWrapper):
    def __init__(self, name, min_included, max_included, null_default_value, **kwargs):
        scipy_distribution = Distribution(
            name=name,
            a=min_included,
            b=max_included,
            **kwargs
        )
        scipy_distribution._pdf = self._pdf

        ScipyContinuousDistributionWrapper.__init__(
            self,
            scipy_distribution=scipy_distribution,
            null_default_value=null_default_value
        )

    def _override_scipy_methods(self):
        self.scipy_distribution._pdf = self._pdf

    @abstractmethod
    def _pdf(self, x, *args):
        pass


class RandInt(ScipyDiscreteDistributionWrapper):
    """
    Rand int scipy distribution. Check out `scipy.stats.randint for more info <https://docs.scipy.org/doc/scipy-0.15.1/reference/generated/scipy.stats.randint.html>`_
    """

    def __init__(self, min_included: int, max_included: int, **kwargs):
        self.min_included = min_included
        self.max_included = max_included
        kwargs.update(low=min_included, high=max_included)

        super().__init__(
            scipy_distribution=randint,
            **kwargs,
        )

    def min(self):
        return self.min_included

    def mean(self, *args, **kwargs):
        return (self.min_included + self.max_included) / 2

    def median(self, *args, **kwargs):
        return self.mean()

    def max(self):
        return self.max_included


class Uniform(BaseCustomContinuousScipyDistribution):
    """
    Get a uniform distribution.

    .. seealso::
        :class:`UniformScipyDistribution`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`neuraxle.base.BaseStep`
    """

    def __init__(self, min_included: float, max_included: float, null_default_value=None):
        """
        Create a random uniform distribution.
        A random float between the two values somehow inclusively will be returned.

        :param min_included: minimum integer, included.
        :type min_included: float
        :param max_included: maximum integer, might be included - for more info, see `examples <https://docs.python.org/2/library/random.html#random.uniform>`__
        :type max_included: float
        :param null_default_value: null default value for distribution. if None, take the min_included
        :type null_default_value: int
        """
        if null_default_value is None:
            null_default_value = min_included

        self.min_included: float = min_included
        self.max_included: float = max_included

        super().__init__(
            name='uniform',
            min_included=min_included,
            max_included=max_included,
            null_default_value=null_default_value
        )

    def _pdf(self, x):
        """
        Calculate the Uniform probability distribution value at position `x`.

        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """
        if self.min_included == self.max_included and (x == self.min_included):
            return 1.

        if (x >= self.min_included) and (x <= self.max_included):
            return 1 / (self.max_included - self.min_included)

        # Manage case where it is outside the probability distribution function.
        return 0.


class LogUniform(BaseCustomContinuousScipyDistribution):
    """
    Get a LogUniform distribution.

    .. seealso::
        :class:`LogUniformScipyDistribution`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`neuraxle.base.BaseStep`
    """

    def __init__(self, min_included: float, max_included: float, null_default_value=None):
        """
        Create a quantized random log uniform distribution.
        A random float between the two values inclusively will be returned.

        :param min_included: minimum integer, should be somehow included.
        :param max_included: maximum integer, should be somehow included.
        :param null_default_value: null default value for distribution. if None, take the min_included
        :type null_default_value: int
        """
        if null_default_value is None:
            null_default_value = math.log2(min_included)
        else:
            null_default_value = math.log2(null_default_value)

        self.min_included: float = min_included
        self.max_included: float = max_included
        self.log2_min_included = math.log2(min_included)
        self.log2_max_included = math.log2(max_included)

        super().__init__(
            name='log_uniform',
            null_default_value=null_default_value,
            min_included=self.log2_min_included,
            max_included=self.log2_max_included
        )

    def _pdf(self, x) -> float:
        """
        Calculate the logUniform probability distribution value at position `x`.

        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """
        if (self.log2_min_included == self.log2_max_included) and x == 2 ** self.log2_min_included:
            return 1.

        if (x >= 2 ** self.log2_min_included) and (x <= 2 ** self.log2_max_included):
            return 1 / (x * math.log(2) * (self.log2_max_included - self.log2_min_included))

        return 0.


class Normal(BaseCustomContinuousScipyDistribution):
    """
    LogNormal distribution that wraps a `continuous scipy distribution <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_continuous.html#scipy.stats.rv_continuous>`_

    Example usage :

    .. code-block:: python

        distribution = LogNormal(
            min_included=0,
            max_included=10,
            null_default_value=0.0
        )

        assert 0.0 <= distribution.rvs() <= 10.0
        assert distribution.pdf(10) < 0.001
        assert distribution.pdf(0) < 0.42
        assert 0.55 > distribution.cdf(5.0) > 0.45
        assert distribution.cdf(0) == 0.0


    .. seealso::
        :class:`LogNormalScipyDistribution`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`neuraxle.base.BaseStep`
    """

    def __init__(
            self,
            mean: float,
            std: float,
            hard_clip_min: float = None,
            hard_clip_max: float = None,
            null_default_value: float = None
    ):
        super().__init__(
            name='normal',
            min_included=hard_clip_min,
            max_included=hard_clip_max,
            null_default_value=null_default_value
        )

        self.hard_clip_min = hard_clip_min
        self.hard_clip_max = hard_clip_max
        self.mean = mean
        self.std = std

    def _pdf(self, x) -> float:
        """
        Calculate the Normal probability distribution value at position `x`.
        :param x: value where the probability distribution function is evaluated.
        :return: value of the probability distribution function.
        """

        if self.hard_clip_min is not None and (x < self.hard_clip_min):
            return 0.

        if self.hard_clip_max is not None and (x > self.hard_clip_max):
            return 0.

        if self.hard_clip_min is not None or self.hard_clip_max is not None:
            a = -np.inf
            b = np.inf

            if self.hard_clip_min is not None:
                a = (self.hard_clip_min - self.mean) / self.std

            if self.hard_clip_max is not None:
                b = (self.hard_clip_max - self.mean) / self.std

            return truncnorm.pdf(x, a=a, b=b, loc=self.mean, scale=self.std)

        return norm.pdf(x, loc=self.mean, scale=self.std)


class LogNormal(BaseCustomContinuousScipyDistribution):
    """
    Get a normal distribution.

    .. seealso::
        :class:`NormalScipyDistribution`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`neuraxle.base.BaseStep`
    """

    def __init__(
            self,
            log2_space_mean: float,
            log2_space_std: float,
            hard_clip_min: float,
            hard_clip_max: float,
            null_default_value: float = None
    ):
        """
        Create a normal distribution from mean and standard deviation.

        :param log2_space_mean: the most common value to pop
        :type log2_space_mean: float
        :param log2_space_std: the standard deviation (that is, the sqrt of the variance).
        :type log2_space_std: float
        :param hard_clip_min: if not none, rvs will return max(result, hard_clip_min).
        :type hard_clip_min: float
        :param hard_clip_max: if not none, rvs will return min(result, hard_clip_min).
        :type hard_clip_max: float
        :param null_default_value: if none, null default value will be set to hard_clip_min.
        :type null_default_value: float
        """
        if null_default_value is None:
            null_default_value = hard_clip_min

        if hard_clip_min is None:
            hard_clip_min = np.nan

        if hard_clip_max is None:
            hard_clip_max = np.nan

        self.log2_space_mean = log2_space_mean
        self.log2_space_std = log2_space_std

        super().__init__(
            name='log_normal',
            min_included=hard_clip_min,
            max_included=hard_clip_max,
            null_default_value=null_default_value
        )

    def _pdf(self, x):
        if x <= 0:
            return 0.

        cdf_min = 0.
        cdf_max = 1.

        pdf_x = 1 / (x * math.log(2) * self.log2_space_std * math.sqrt(2 * math.pi)) * math.exp(
            -(math.log2(x) - self.log2_space_mean) ** 2 / (2 * self.log2_space_std ** 2))
        return pdf_x / (cdf_max - cdf_min)


class Gaussian(BaseCustomContinuousScipyDistribution):
    """
    Gaussian distribution that inherits from `scipy.stats.rv_continuous <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_continuous.html#scipy.stats.rv_continuous>`_

    Example usage :

    .. code-block:: python

        gaussian_distribution = Gaussian(
            min_included=0,
            max_included=10,
            null_default_value=0.0
        )

        assert 0.0 <= gaussian_distribution.rvs() <= 10.0
        assert gaussian_distribution.pdf(10) < 0.001
        assert gaussian_distribution.pdf(0) < 0.42
        assert 0.55 > gaussian_distribution.cdf(5.0) > 0.45
        assert gaussian_distribution.cdf(0) == 0.0


    .. seealso::
        :class:`GaussianScipyDistribution`,
        :class:`ScipyDistributionWrapper`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`HyperparameterDistribution`,
        :class:`neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`neuraxle.base.BaseStep`
    """

    def __init__(self, min_included: int, max_included: int, null_default_value: float = None):
        self.max_included = max_included
        self.min_included = min_included

        BaseCustomContinuousScipyDistribution.__init__(
            self,
            name='gaussian',
            min_included=min_included,
            max_included=max_included,
            null_default_value=null_default_value
        )

    def _pdf(self, x):
        return math.exp(-x ** 2 / 2.) / np.sqrt(2.0 * np.pi)


class Poisson(BaseCustomDiscreteScipyDistribution):
    """
    Poisson distribution that inherits from `scipy.stats.rv_discrete <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_histogram.html#scipy.stats.rv_histogram>`_

    Example usage :

    .. code-block:: python

        poisson_distribution = Poisson(
            min_included=0.0,
            max_included=10.0,
            null_default_value=0.0,
            mu=5.0
        )

        rvs = [poisson_distribution.rvs() for i in range(10)]
        assert not all(x == rvs[0] for x in rvs)
        assert 0.0 <= poisson_distribution.rvs() <= 10.0
        assert poisson_distribution.pdf(10) == 0.01813278870782187
        assert np.isclose(poisson_distribution.pdf(0), 0.006737946999085467)
        assert poisson_distribution.cdf(5.0) == 0.6159606548330632
        assert poisson_distribution.cdf(0) == 0.006737946999085467


    .. seealso::
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`PoissonScipyDistribution`,
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, min_included: float, max_included: float, null_default_value: float = None, mu=0.6):
        super().__init__(
            min_included=min_included,
            max_included=max_included,
            name='poisson',
            null_default_value=null_default_value
        )
        self.mu = mu

    def _pmf(self, x):
        return math.exp(-self.mu) * self.mu ** x / factorial(x)


class Histogram(ScipyDistributionWrapper):
    """
    Histogram distribution that inherits from `scipy.stats.rv_histogram <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.rv_histogram.html#scipy.stats.rv_histogram>`_

    Example usage :

    .. code-block:: python

        hist_dist = Histogram(
             histogram=np.histogram(data, bins=100),
             null_default_value=0.0
        )

        assert min(data) <= hist_dist.rvs() <= max(data)
        assert 1.0 > hist_dist.pdf(x=1.0) > 0.0
        assert hist_dist.pdf(x=np.max(data)) == 0.0
        assert hist_dist.pdf(x=np.min(data)) < 0.001
        assert hist_dist.cdf(x=np.max(data)) == 1.0
        assert 0.55 > hist_dist.cdf(x=np.median(data)) > 0.45
        assert hist_dist.cdf(x=np.min(data)) == 0.0


    .. seealso::
        :class:`HyperparameterDistribution`,
        :class:`ScipyDistributionWrapper`,
        :class:`Poisson`,
        :class:`Gaussian`,
        :func:`~neuraxle.base.BaseStep.set_hyperparams_space`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, histogram: np.histogram, null_default_value: float = None, **kwargs):
        ScipyDistributionWrapper.__init__(
            self,
            scipy_distribution=rv_histogram(histogram=histogram, **kwargs)
        )
        ContinuousHyperparameterDistrbution.__init__(self, null_default_value=null_default_value)


class FixedHyperparameter(BaseCustomDiscreteScipyDistribution):
    """This is an hyperparameter that won't change again, but that is still expressed as a distribution."""

    def __init__(self, value, null_default_value=None):
        """
        Create a still hyperparameter

        :param value: what will be returned by calling ``.rvs()``.
        """
        self.value = value

        super().__init__(
            name='fixed',
            min_included=value,
            max_included=value,
            null_default_value=null_default_value
        )

    def _pdf(self, x):
        if x == self.value:
            return 1.
        return 0.


class Boolean(BaseCustomDiscreteScipyDistribution):
    """Get a random boolean hyperparameter."""

    def __init__(self, proba_is_true: Optional[float] = None, null_default_value=False):
        """
        Create a boolean hyperparameter with given probability.

        Boolean distribution is in fact a Bernouilli distribution where given probability
        set occurance probability of having value 1. (1 - probability) gives occurance probability of having value 0.

        :param proba: a float corresponding to proportion of 1 over 0.
        :type proba: float
        :param null_default_value: default value for distribution
        :type null_default_value: default choice value. if None, default choice value will be the first choice
        """
        if proba_is_true is None:
            proba_is_true = 0.5

        if not (0 <= proba_is_true <= 1):
            raise ValueError("Probability must be between 0 and 1 (inclusively).")

        self.proba_is_true = proba_is_true

        super().__init__(
            name='boolean',
            min_included=0,
            max_included=1,
            null_default_value=null_default_value
        )

    def _pmf(self, x):
        """
        Calculate the boolean probability mass function value at position `x`.
        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """
        if (x is True) or (x == 1):
            return self.proba_is_true

        if (x is False) or (x == 0):
            return 1 - self.proba_is_true

        return 0.


def get_index_in_list_with_bool(choice_list: List[Any], value: Any) -> int:
    for choice_index, choice in enumerate(choice_list):
        if choice == value and not isinstance(choice, bool) and not isinstance(value, bool):
            index = choice_index
            return index

        if choice is value:
            index = choice_index
            return index

    raise ValueError("{} is not in list".format(value))


class Choice(BaseCustomDiscreteScipyDistribution):
    """Get a random value from a choice list of possible value for this hyperparameter.

    When narrowed, the choice will only collapse to a single element when narrowed enough.
    For example, if there are 4 items in the list, only at a narrowing value of 0.25 that
    the first item will be kept alone.
    """

    def __init__(self, choice_list: List, probas: Optional[List[float]] = None, null_default_value=None):
        """
        Create a random choice hyperparameter from the given list.

        :param choice_list: a list of values to sample from.
        :type choice_list: List
        :param null_default_value: default value for distribution
        :type null_default_value: default choice value. if None, default choice value will be the first choice
        """
        self.choice_list = choice_list

        if probas is None:
            probas = [1 / len(self.choice_list) for _ in self.choice_list]

        # Normalize probas juste in case sum is not equal to one.
        self.probas = np.array(probas) / np.sum(probas)

        if null_default_value is None:
            null_default_value = choice_list[0]
        elif null_default_value in choice_list:
            null_default_value = null_default_value
        else:
            raise ValueError(
                'invalid default value {0} not in choice list : {1}'.format(null_default_value, choice_list))

        super().__init__(
            name='choice',
            min_included=0,
            max_included=len(choice_list) - 1,
            null_default_value=null_default_value
        )

    def probabilities(self):
        return self.probas

    def values(self):
        return self.choice_list

    def rvs(self, *args, **kwargs):
        sample_choice_index = super().rvs(*args, **kwargs)
        if isinstance(sample_choice_index, np.ndarray):
            return [self.choice_list[int(i)] for i in sample_choice_index]
        return self.choice_list[int(sample_choice_index)]

    def pdf(self, x, *args, **kwargs):
        choice_index = [str(choice) for choice in self.choice_list].index(str(x))
        return self.scipy_distribution.pmf(choice_index, *args, **kwargs, **self.scipy_distribution_arguments)

    def cdf(self, x, *args, **kwargs):
        try:
            index = [str(choice) for choice in self.choice_list].index(str(x))
        except ValueError:
            raise ValueError(
                "Item not found in list. Make sure the item is in the choice list and a correct method  __eq__ is defined for all item in the list.")
        except (NotImplementedError, NotImplemented):
            raise ValueError("A correct method for __eq__ should be defined for all item in the list.")
        except AttributeError:
            raise ValueError("choice_list param should be a list.")
        else:
            probas = np.array(self.probas)
            return np.sum(probas[0:index + 1])

    def mean(self):
        """
        Calculate mean value (also called esperance) of the random variable.
        :return: mean value of the random variable.
        """
        choice_index = np.arange(0, len(self), 1)
        probas = np.array(self.probas)
        mean = np.sum(choice_index * probas)
        return mean

    def var(self):
        """
        Calculate variance value of the random variable.
        :return: variance value of the random variable.
        """
        choice_index = np.arange(0, len(self), 1)
        probas = np.array(self.probas)
        mean = np.sum(choice_index * probas)
        second_moment = np.sum(choice_index ** 2 * probas)
        var = second_moment - mean ** 2
        return var

    def min(self):
        return 0

    def max(self):
        return len(self.choice_list) - 1

    def std(self):
        return np.std([i for i, _ in enumerate(self.choice_list)])

    def pmf(self, x):
        """
        Calculate the choice probability mass function value at position `x`.

        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """
        if len(self.choice_list) - 1 >= x[-1] >= 0:
            return sum([self.probas[int(i)] for i in x])
        else:
            return 0.

    def _pmf(self, x):
        """
        Calculate the choice probability mass function value at position `x`.

        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """
        if len(self.choice_list) - 1 >= x[-1] >= 0:
            return sum([self.probas[int(i)] for i in x])
        else:
            return 0.

    def __len__(self):
        return len(self.choice_list)


class Quantized(WrappedHyperparameterDistributions, BaseCustomContinuousScipyDistribution):
    """A quantized wrapper for another distribution: will round() the rvs number."""

    def __init__(self, hd: HyperparameterDistribution = None, hds: List[HyperparameterDistribution] = None,
                 null_default_value=None):
        WrappedHyperparameterDistributions.__init__(
            self,
            hd=hd,
            hds=hds,
            null_default_value=null_default_value
        )

        hd_min = self.hd.min()
        if np.isneginf(hd_min):
            min_included = hd_min
        else:
            min_included = round(hd_min)

        hd_max = self.hd.max()
        if np.isposinf(hd_max):
            max_included = hd_max
        else:
            max_included = round(hd_max)

        BaseCustomContinuousScipyDistribution.__init__(
            self,
            name='quantized',
            min_included=min_included,
            max_included=max_included,
            null_default_value=self.null_default_value,
            is_continuous=True
        )

    def _pdf(self, x):
        """
        Calculate the Quantized probability mass function value at position `x` of a continuous distribution.
        :param x: value where the probability mass function is evaluated.
        :return: value of the probability mass function.
        """
        # In order to calculate the pdf for any quantized distribution,
        # we have to perform the integral from x-0.5 to x+0.5 (because of round).
        if isinstance(x, int) or (isinstance(x, float) and x.is_integer()):
            return quad(self.hd.pdf, x - 0.5, x + 0.5)[0]
        return 0.



"""
Neuraxle's metrics classes
=================================================
The neuraxle classes to track metrics results.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
from typing import Dict

from neuraxle.base import MetaStep, BaseStep, ExecutionContext, _FittableStep
from neuraxle.data_container import DataContainer
from neuraxle.hyperparams.space import RecursiveDict


class MetricsWrapper(MetaStep):
    """
    Add metrics calculation to a step. Calculates metrics after each fit, fit_transform, or even transform if there is an expected outputs.

    Example usage :

    .. code-block:: python

        wrapped = MetricsWrapper(wrapped=wrapped, metrics=self.batch_metrics, name=BATCH_METRICS_STEP_NAME)
        wrapped = MiniBatchSequentialPipeline(
            [wrapped],
            batch_size=self.batch_size,
            enabled=True
        )

        # toggle metrics on, and off
        wrapped.apply('toggle_metrics')


    .. seealso::
        :class:`~neuraxle.api.DeepLearningPipeline`,
        :class:`~neuraxle.base.MetaStepMixin`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(
            self,
            wrapped: BaseStep,
            metrics: Dict,
            name: str = None,
            print_metrics=False,
            print_fun=print
    ):
        MetaStep.__init__(self, wrapped, name=name)

        self.metrics: Dict = metrics
        self._initialize_metrics(metrics)

        self.print_metrics = print_metrics
        self.print_fun = print_fun
        self.enabled = True

    def _initialize_metrics(self, metrics):
        """
        Initialize metrics results dict for train, and validation using the metrics function dict.

        :param metrics: metrics function dict
        :type metrics: dict
        :return:
        """
        self.metrics_results_train = {}
        self.metrics_results_validation = {}

        for m in metrics:
            self.metrics_results_train[m] = []
            self.metrics_results_validation[m] = []

    def _did_fit_transform(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        return self._did_fit_transform_or_transform(data_container, context)

    def _did_transform(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        return self._did_fit_transform_or_transform(data_container, context)

    def _did_fit_transform_or_transform(self, data_container: DataContainer, context: ExecutionContext):
        """
        Calculate metrics results after fit, or transform if there is an expected outputs in the data container.
        Also, calculate validation metrics if there is a sub data container named validation in the data container.
        Please refer to :class:`~neuraxle.data_container.DataContainer` for more information about sub data containers.

        :param data_container: data container to calculate metrics for
        :type data_container: DataContainer
        :return: data container
        :rtype: DataContainer
        """
        if data_container.expected_outputs is None or len(data_container.expected_outputs) == 0:
            return data_container

        self._calculate_metrics_results(data_container)

        if 'validation' in data_container and self.enabled:
            self.set_train(False)

            self.apply('disable_metrics')
            validation_data_container = self._transform_data_container(data_container['validation'].copy(), context)
            self.apply('enable_metrics')
            self._calculate_metrics_results(validation_data_container)

            self.set_train(True)

        return data_container

    def _calculate_metrics_results(self, data_container: DataContainer):
        """
        Calculate metrics results using the transformed data container, and the metrics function dict.

        :param data_container: transformed data container
        :type data_container: DataContainer
        :return:
        """
        if not self.enabled:
            return

        result = {}
        for metric_name, metric_function in self.metrics.items():
            result_metric = metric_function(data_container.data_inputs, data_container.expected_outputs)
            result[metric_name] = result_metric

            if self.is_train:
                self.metrics_results_train[metric_name].append(result_metric)
            else:
                self.metrics_results_validation[metric_name].append(result_metric)

        if self.print_metrics:
            self.print_fun(result)

    def get_metrics(self) -> RecursiveDict:
        """
        Get all metrics results using the transformed data container, and the metrics function dict.
        To be used with :func:`neuraxle.base.BaseStep.apply` method.

        :return: dict with the step name as key, and all of the training, and validation metrics as values
        """
        return RecursiveDict({
            'train': self.metrics_results_train,
            'validation': self.metrics_results_validation
        })

    def toggle_metrics(self) -> RecursiveDict:
        """
        Toggle metrics wrapper on and off to temporarily disable metrics if needed..

        :return:
        """
        self.enabled = not self.enabled
        return RecursiveDict({self.name: self.enabled})

    def disable_metrics(self):
        """
        Disable metrics wrapper metrics if needed..

        :return:
        """
        self.enabled = False
        return RecursiveDict({self.name: self.enabled})

    def enable_metrics(self):
        """
        Enable metrics wrapper metrics if needed..

        :return:
        """
        self.enabled = True
        return RecursiveDict({self.name: self.enabled})



"""
Pipeline Steps Based on NumPy
=====================================
Those steps works with NumPy (np) arrays.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import numpy as np

from neuraxle.base import DataContainer, ExecutionContext, ForceHandleMixin, BaseTransformer
from neuraxle.base import NonFittableMixin, BaseStep
from neuraxle.hyperparams.space import HyperparameterSamples


class NumpyFlattenDatum(BaseTransformer):
    def __init__(self):
        BaseTransformer.__init__(self)

    def transform(self, data_inputs):
        return data_inputs.reshape(data_inputs.shape[0], -1)


class NumpyConcatenateOnCustomAxis(BaseTransformer):
    """
    Numpy concetenation step where the concatenation is performed along the specified custom axis.
    """

    def __init__(self, axis):
        """
        Create a numpy concatenate on custom axis object.
        :param axis: the axis where the concatenation is performed.
        :return: NumpyConcatenateOnCustomAxis instance.
        """
        self.axis = axis
        BaseTransformer.__init__(self)

    def _transform_data_container(self, data_container, context):
        """
        Handle transform.

        :param data_container: the data container to join
        :param context: execution context
        :return: transformed data container
        """
        data_inputs = self.transform([dc.data_inputs for dc in data_container.data_inputs])
        data_container = DataContainer(data_inputs=data_inputs, current_ids=data_container.current_ids,
                                       expected_outputs=data_container.expected_outputs)
        data_container.set_data_inputs(data_inputs)

        return data_container

    def transform(self, data_inputs):
        """
        Apply the concatenation transformation along the specified axis.
        :param data_inputs:
        :return: Numpy array
        """
        return self._concat(data_inputs)

    def _concat(self, data_inputs):
        return np.concatenate(data_inputs, axis=self.axis)


class NumpyConcatenateInnerFeatures(NumpyConcatenateOnCustomAxis):
    """
    Numpy concatenation step where the concatenation is performed along `axis=-1`.
    """

    def __init__(self):
        """
        Create a numpy concatenate inner features object.
        :return: NumpyConcatenateOnCustomAxis instance.
        """
        # The concatenate is on the inner features so axis = -1.
        NumpyConcatenateOnCustomAxis.__init__(self, axis=-1)


class NumpyConcatenateOuterBatch(NumpyConcatenateOnCustomAxis):
    """
    Numpy concetenation step where the concatenation is performed along `axis=0`.
    """

    def __init__(self):
        """
        Create a numpy concatenate outer batch object.
        :return: NumpyConcatenateOnCustomAxis instance which is inherited by base step.
        """
        NumpyConcatenateOnCustomAxis.__init__(self, axis=0)


class NumpyTranspose(BaseTransformer):
    def __init__(self):
        super().__init__()

    def _transform_data_container(self, data_container, context):
        """
        Handle transform.

        :param data_container: the data container to join
        :param context: execution context
        :return: transformed data container
        """
        data_inputs = self.transform([dc.data_inputs for dc in data_container.data_inputs])
        data_container = DataContainer(data_inputs=data_inputs, current_ids=data_container.current_ids,
                                       expected_outputs=data_container.expected_outputs)
        data_container.set_data_inputs(data_inputs)

        return data_container

    def transform(self, data_inputs):
        return self._transpose(data_inputs)

    def inverse_transform(self, data_inputs):
        return self._transpose(data_inputs)

    def _transpose(self, data_inputs):
        return np.array(data_inputs).transpose()


class NumpyShapePrinter(BaseTransformer):

    def __init__(self, custom_message: str = ""):
        super().__init__()
        self.custom_message = custom_message

    def transform(self, data_inputs):
        self._print(data_inputs)
        return data_inputs

    def inverse_transform(self, processed_outputs):
        self._print(processed_outputs)
        return processed_outputs

    def _print(self, data_inputs):
        print(self.__class__.__name__ + ":", data_inputs.shape, self.custom_message)

    def _print_one(self, data_input):
        print(self.__class__.__name__ + " (one):", data_input.shape, self.custom_message)


class MultiplyByN(BaseTransformer):
    """
    Step to multiply a numpy array.
    Accepts an integer for the number to multiply by.

    Example usage:

    .. code-block:: python

        pipeline = Pipeline([
            MultiplyByN(3)
        ])
        outputs = pipeline.transform(np.array([1])
        # outputs => np.array([3])

    .. seealso::
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, multiply_by=1):
        super().__init__(hyperparams=HyperparameterSamples({ 'multiply_by': multiply_by }))

    def transform(self, data_inputs):
        if not isinstance(data_inputs, np.ndarray):
            data_inputs = np.array(data_inputs)

        return data_inputs * self.hyperparams['multiply_by']

    def inverse_transform(self, data_inputs):
        if not isinstance(data_inputs, np.ndarray):
            data_inputs = np.array(data_inputs)

        return data_inputs / self.hyperparams['multiply_by']


class AddN(BaseTransformer):
    """
    Step to add a scalar to a numpy array.
    Accepts an integer for the number to add to every data inputs.

    Example usage:

    .. code-block:: python

        pipeline = Pipeline([
            AddN(1)
        ])
        outputs = pipeline.transform(np.array([1])
        # outputs => np.array([2])

    .. seealso::
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(self, add=1):
        super().__init__(hyperparams=HyperparameterSamples({ 'add': add }))

    def transform(self, data_inputs):
        if not isinstance(data_inputs, np.ndarray):
            data_inputs = np.array(data_inputs)

        return data_inputs + self.hyperparams['add']

    def inverse_transform(self, data_inputs):
        if not isinstance(data_inputs, np.ndarray):
            data_inputs = np.array(data_inputs)

        return data_inputs - self.hyperparams['add']


class Sum(BaseTransformer):
    """
    Step sum numpy array using np.sum.

    Example usage:

    .. code-block:: python

        pipeline = Pipeline([
            Sum(axis=-1)
        ])

        outputs = pipeline.transform(np.array([1, 2, 3])
        # outputs => 6)

    .. seealso::
        :class:`BaseStep`
    """

    def __init__(self, axis):
        BaseTransformer.__init__(self)
        self.axis = axis

    def transform(self, data_inputs):
        if not isinstance(data_inputs, np.ndarray):
            data_inputs = np.array(data_inputs)

        data_inputs = np.expand_dims(np.sum(data_inputs, axis=self.axis), axis=-1)
        return data_inputs


class OneHotEncoder(BaseTransformer):
    """
    Step to one hot a set of columns.
    Accepts Integer Columns and converts it ot a one_hot.
    Rounds floats  to integer for safety in the transform.
    
    Example usage: 
    
    1. Set up data

    .. code-block:: python

       import numpy as np
       a = np.array([1,0,3])
       b = np.array([[0,1,0,0], [1,0,0,0], [0,0,0,1]])

    2. **Do the actual conversion**

    .. code-block:: python

       from neuraxle.steps.numpy import OneHotEncoder
       encoder = OneHotEncoder(nb_columns=4)
       b_pred = encoder.transform(a)

    3. Assert it works

    .. code-block:: python

       assert b_pred == b
    
    .. seealso::
        `StackOverflow answer about one hot encoding <https://stackoverflow.com/a/59262363/2476920>`__
    
    """

    def __init__(self, nb_columns, name):
        super().__init__(name=name)
        self.nb_columns = nb_columns

    def transform(self, data_inputs):
        """
        Transform data inputs using one hot encoding, adding no_columns to the -1 axis.
        :param data_inputs: data inputs to encode
        :return: one hot encoded data inputs
        """
        # validate enum values
        if not isinstance(data_inputs, np.ndarray):
            data_inputs = np.array(data_inputs)

        # treats invalid values as having no columns activated. create a temporary column for invalid values
        data_inputs[data_inputs == None] = self.nb_columns
        data_inputs[data_inputs >= self.nb_columns] = self.nb_columns
        data_inputs[data_inputs < 0] = self.nb_columns

        # finally, one hot encode data inputs
        outputs_ = np.eye(self.nb_columns + 1)[np.array(data_inputs, dtype=np.int32)]

        # delete the invalid values column, and zero hot the invalid values
        outputs_ = np.delete(outputs_, self.nb_columns, axis=-1)

        return outputs_.squeeze()


class ToNumpy(ForceHandleMixin, BaseTransformer):
    """
    Convert data inputs, and expected outputs to a numpy array.
    """

    def _will_process(self, data_container: DataContainer, context: ExecutionContext) -> (
            DataContainer, ExecutionContext):
        return data_container.to_numpy(), context


class NumpyReshape(BaseTransformer):
    """
    Reshape numpy array in data inputs.

    .. code-block:: python

       import numpy as np
       a = np.array([1,0,3])
       outputs = NumpyReshape(shape=(-1,1)).transform(a)
       assert np.array_equal(outputs, np.array([[1],[0],[3]]))

    .. seealso::
        :class:`BaseStep`
    """

    def __init__(self, new_shape):
        super().__init__()
        self.new_shape = new_shape

    def transform(self, data_inputs):
        return np.reshape(data_inputs, newshape=self.new_shape)


class NumpyRavel(NonFittableMixin, BaseStep):
    """
    Return a contiguous flattened array using `np.ravel <https://numpy.org/doc/stable/reference/generated/numpy.ravel.html>̀_

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`
    """

    def transform(self, data_inputs):
        """
        Flatten numpy array using ̀np.ravel <https://numpy.org/doc/stable/reference/generated/numpy.ravel.html>`_

        :param data_inputs:
        :return:
        """
        if data_inputs is None:
            return data_inputs

        data_inputs = data_inputs if isinstance(data_inputs, np.ndarray) else np.array(data_inputs)
        return data_inputs.ravel()


class NumpyFFT(NonFittableMixin, BaseStep):
    """
    Compute time series FFT using `np.fft.rfft <https://numpy.org/doc/stable/reference/generated/numpy.fft.rfft.html>`_

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`
    """

    def __init__(self, axis=None):
        BaseStep.__init__(self)
        NonFittableMixin.__init__(self)
        if axis is None:
            axis = -2
        self.axis = axis

    def transform(self, data_inputs):
        """
        Will featurize time series data with FFT using `np.fft.rfft <>̀_


        :param data_inputs: time series data of 3D shape: [batch_size, time_steps, sensors_readings]
        :return: featurized data is of 2D shape: [batch_size, n_features]
        """
        return np.fft.rfft(data_inputs, axis=self.axis)


class NumpyAbs(NonFittableMixin, BaseStep):
    """
    Compute `np.abs <https://numpy.org/doc/stable/reference/generated/numpy.absolute.html>`_.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`
    """

    def transform(self, data_inputs):
        """
        Will featurize data with a absolute value transformation.

        :param data_inputs: data inputs
        :return: absolute data inputs
        """
        return np.abs(data_inputs)


class NumpyMean(NonFittableMixin, BaseStep):
    """
    Compute `np.mean <https://numpy.org/devdocs/reference/generated/numpy.mean.html>`_ at the given axis.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`
    """

    def __init__(self, axis=None):
        BaseStep.__init__(self)
        NonFittableMixin.__init__(self)

        if axis is None:
            axis = -2
        self.axis = axis

    def transform(self, data_inputs):
        """
        Will featurize data with a mean.

        :param data_inputs: data inputs
        :return: data inputs mean for the given axis
        """
        return np.mean(data_inputs, axis=self.axis)


class NumpyMedian(NonFittableMixin, BaseStep):
    """
    Compute `np.median <https://numpy.org/doc/1.18/reference/generated/numpy.median.html>`_ at the given axis.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`
    """

    def __init__(self, axis=None):
        BaseStep.__init__(self)
        NonFittableMixin.__init__(self)

        if axis is None:
            axis = -2
        self.axis = axis

    def transform(self, data_inputs):
        """
        Will featurize data with a median.

        :param data_inputs: data inputs
        :return: data inputs median for the given axis
        """
        return np.median(data_inputs, axis=self.axis)


class NumpyMin(NonFittableMixin, BaseStep):
    """
    Compute `np.min <https://numpy.org/doc/1.18/reference/generated/numpy.minimum.html>`_ at the given axis.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`
    """

    def __init__(self, axis=None):
        BaseStep.__init__(self)
        NonFittableMixin.__init__(self)

        if axis is None:
            axis = -2
        self.axis = axis

    def transform(self, data_inputs):
        """
        Will featurize data with a min.

        :param data_inputs: data inputs
        :return: min value for the given axis
        """
        return np.min(data_inputs, axis=self.axis)


class NumpyMax(NonFittableMixin, BaseStep):
    """
    Compute `np.max <https://numpy.org/doc/1.18/reference/generated/numpy.ndarray.max.html>̀_ at the given axis.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`
    """

    def __init__(self, axis=None):
        BaseStep.__init__(self)
        NonFittableMixin.__init__(self)

        if axis is None:
            axis = -2
        self.axis = axis

    def transform(self, data_inputs):
        """
        Will featurize data with a max.

        :param data_inputs: 3D time series of shape [batch_size, time_steps, sensors]
        :return: max value for the given axis
        """
        return np.max(data_inputs, axis=self.axis)


class NumpyArgMax(NonFittableMixin, BaseStep):
    """
    Compute `np.max <https://numpy.org/doc/1.18/reference/generated/numpy.ndarray.argmax.html>̀_ at the given axis.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.NonFittableMixin`
    """

    def __init__(self, axis=None):
        BaseStep.__init__(self)
        NonFittableMixin.__init__(self)

        if axis is None:
            axis = -2
        self.axis = axis

    def transform(self, data_inputs):
        """
        Will featurize data with a max.

        :param data_inputs: 3D time series of shape [batch_size, time_steps, sensors]
        :return: max value for the given axis
        """
        return np.argmax(data_inputs, axis=self.axis)





"""
Notebook matplotlib plotting functions
================================================

Utility function for plotting in notebooks.

..
    Note: some of the code in the present code block is derived from another project licensed under The MIT License (MIT),
    Copyright (c) 2017 Vooban Inc. For the full information, see:
    https://github.com/guillaume-chevalier/Hyperopt-Keras-CNN-CIFAR-100/blob/Vooban/LICENSE

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import matplotlib.pyplot as plt
import os
from neuraxle.metaopt.auto_ml import HyperparamsRepository

from neuraxle.hyperparams.distributions import *
from neuraxle.hyperparams.space import HyperparameterSpace
from neuraxle.metaopt.observable import _Observer, T
from neuraxle.metaopt.trial import Trial, TRIAL_STATUS, Trials

DISCRETE_NUM_BINS = 40
CONTINUOUS_NUM_BINS = 1000
NUM_TRIALS = 100000
X_DOMAIN = np.array(range(-100, 600)) / 100


def plot_histogram(title: str, distribution: HyperparameterDistribution, num_bins=50):
    samples = np.array([
        distribution.rvs() for _ in range(NUM_TRIALS)
    ], dtype=np.float).flatten()

    pdf_values = []
    for value in X_DOMAIN:
        try:
            pdf_values.append(distribution.pdf(value))
        except ValueError:
            pdf_values.append(0.)

    pdf_values = np.array(pdf_values)

    plt.figure(figsize=(17, 5))
    ax = plt.gca()
    hist = plt.hist(samples, bins=num_bins, label="hist")
    renormalization_const = hist[0].max() / pdf_values.max()
    renormalized_pdf_values = renormalization_const * pdf_values

    plt.plot(X_DOMAIN, renormalized_pdf_values, label="renorm pdf")
    plt.title("Histogram (pdf) for a {} distribution".format(title))
    handles, labels = ax.get_legend_handles_labels()
    # sort both labels and handles by labels
    labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
    ax.legend(handles, labels)
    plt.xlim([X_DOMAIN.min(), X_DOMAIN.max()])
    plt.show()


def plot_pdf_cdf(title: str, distribution: HyperparameterDistribution):
    pdf_values = []
    cdf_values = []
    for value in X_DOMAIN:
        try:
            pdf_values.append(distribution.pdf(value))
        except ValueError:
            pdf_values.append(0.)

        try:
            cdf_values.append(distribution.cdf(value))
        except ValueError:
            cdf_values.append(0.)

    pdf_values = np.array(pdf_values)
    cdf_values = np.array(cdf_values)
    plt.figure(figsize=(17, 5))
    ax = plt.gca()
    plt.plot(X_DOMAIN, pdf_values, label="pdf")
    plt.plot(X_DOMAIN, cdf_values, label="cdf")
    plt.title("Pdf and cdf for a {} distribution".format(title))
    handles, labels = ax.get_legend_handles_labels()
    # sort both labels and handles by labels
    labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
    ax.legend(handles, labels)
    plt.xlim([X_DOMAIN.min(), X_DOMAIN.max()])
    plt.show()


def plot_distribution_space(hyperparameter_space: HyperparameterSpace, num_bins=50):
    for title, distribution in hyperparameter_space.items():
        print(title + ":")
        plot_histogram(title, distribution, num_bins=num_bins)
        plot_pdf_cdf(title, distribution)


class TrialMetricsPlottingObserver(_Observer[Tuple[HyperparamsRepository, Trial]]):
    def __init__(
            self,
            plotting_folder_name: str = 'metric_results',
            save_plots: bool = True,
            plot_trial_on_next: bool = True,
            plot_all_trials_on_complete: bool = True,
            plot_individual_trials_on_complete: bool = True
    ):
        self.plot_individual_trials_on_complete = plot_individual_trials_on_complete
        self.plot_trial_on_next: bool = plot_trial_on_next
        self.plot_all_trials_on_complete: bool = plot_all_trials_on_complete
        self.plotting_folder_name: str = plotting_folder_name
        self.save: bool = save_plots

    def on_next(self, value: Tuple[HyperparamsRepository, Trial]):
        repo, trial = value
        if not self.plot_trial_on_next:
            return

        self._plot_all_trial_main_and_validation_metric_results(repo, trial)

    def _plot_all_trial_main_and_validation_metric_results(self, repo, trial):
        trial_hash = repo._get_trial_hash(trial)
        for split in trial.validation_splits:
            for metric_name in split.get_metric_names():
                train_results = split.get_metric_train_results(metric_name=metric_name)
                validation_results = split.get_metric_validation_results(metric_name=metric_name)
                plt.plot(train_results)
                plt.plot(validation_results)
                plt.ylabel(metric_name)
                plt.xlabel('epoch')
                plt.legend(['train', 'validation'], loc='upper left')
                plt.title(metric_name)
                plotting_file = os.path.join(repo.cache_folder, trial_hash, str(split), '{}.png'.format(metric_name))
                self._show_or_save_plot(plotting_file)

    def on_complete(self, value: Tuple[HyperparamsRepository, Trial]):
        repo, trial = value
        trials: Trials = repo.load_all_trials(TRIAL_STATUS.SUCCESS)
        if not self.plot_all_trials_on_complete:
            return
        if len(trials) == 0:
            return

        if self.plot_individual_trials_on_complete:
            for trial in trials:
                self._plot_all_trial_main_and_validation_metric_results(repo, trial)

        if self.plot_all_trials_on_complete:
            self._plot_all_trials_on_complete(repo, trials)

    def _plot_all_trials_on_complete(self, repo, trials):
        n_splits = trials.get_number_of_split()
        metric_names = trials.get_metric_names()
        for metric_name in metric_names:
            for split_number in range(n_splits):
                self._plot_all_trials_training_results_for_metric(
                    trials=trials,
                    metric_name=metric_name,
                    cache_folder=repo.cache_folder,
                    split_number=split_number
                )
                self._plot_all_trials_validation_results_for_metric(
                    trials=trials,
                    metric_name=metric_name,
                    cache_folder=repo.cache_folder,
                    split_number=split_number
                )

    def _plot_all_trials_validation_results_for_metric(self, trials, metric_name, cache_folder, split_number):
        for trial in trials:
            validation_results = trial[split_number].get_metric_validation_results(metric_name=metric_name)
            plt.plot(validation_results)

        plt.ylabel(metric_name)
        plt.xlabel('epoch')
        plt.title('Validation {}'.format(metric_name))
        metric_file_name = '{}_{}_validation.png'.format(metric_name, split_number)
        plotting_file = os.path.join(cache_folder, self.plotting_folder_name, str(split_number), metric_file_name)
        self._show_or_save_plot(plotting_file)

    def _plot_all_trials_training_results_for_metric(self, trials, metric_name, cache_folder, split_number):
        for trial in trials:
            validation_results = trial[split_number].get_metric_train_results(metric_name=metric_name)
            plt.plot(validation_results, linewidth=0.5)

        plt.ylabel(metric_name)
        plt.xlabel('epoch')
        plt.title('Train {}'.format(metric_name))
        metric_file_name = '{}_{}_train.png'.format(metric_name, split_number)
        plotting_file = os.path.join(cache_folder, self.plotting_folder_name, str(split_number), metric_file_name)
        self._show_or_save_plot(plotting_file)

    def _show_or_save_plot(self, plotting_file):
        if self.save:
            plt.savefig(plotting_file)
        else:
            plt.show()
        plt.close()



"""
Neuraxle
====================================
This is the core of Neuraxle. Most pipeline steps derive (inherit) from those classes. They are worth noticing.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

from setuptools import setup, find_packages

from neuraxle import __version__ as _VERSION

with open('README.rst') as _f:
    _README = _f.read()

setup(
    name='neuraxle',
    version=_VERSION,
    description='Neuraxle is a Machine Learning (ML) library for building neat pipelines, providing the right '
                'abstractions to both ease research, development, and deployment of your ML applications.',
    long_description=_README,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        'License :: OSI Approved :: Apache Software License',
        "Natural Language :: English",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "Topic :: Adaptive Technologies",
        "Topic :: Office/Business",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Artificial Life",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development",
        "Topic :: Software Development :: Assemblers",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Object Brokering",
        "Topic :: Software Development :: Pre-processors",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
        "Topic :: System",
        "Topic :: System :: Clustering",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Networking",
        # Topic :: System :: Systems Administration,
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Filters",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Utilities",
        "Typing :: Typed"
    ],
    url='https://github.com/Neuraxio/Neuraxle',
    download_url='https://github.com/Neuraxio/Neuraxle/tarball/{}'.format(
        _VERSION),
    author='Neuraxio Inc.',
    author_email='guillaume.chevalier@neuraxio.com',
    packages=find_packages(include=['neuraxle*']),
    test_suite="testing",
    setup_requires=["pytest-runner"],
    install_requires=[
        'numpy',
        'scipy>=1.4.1',
        'matplotlib',
        'joblib>=0.13.2',
        'Flask>=1.1.1',
        'Flask-RESTful>=0.3.7',
        'conv==0.2'
    ],
    tests_require=[
        "pytest",
        "pytest-cov",
        "scikit-learn>=0.20.3"
    ],
    include_package_data=True,
    license='Apache 2.0',
    keywords='pipeline pipelines data science machine learning deep learning'
)

print("""
____________________________________________________________________

 Thank you for installing
   _   _                               __
  | \ | |                             |  |
  |  \| | ___  _   _  _ __  ___ __  __ | |  ___
  | . ` |/ _ \| | | || ' _||__ \\\\ \/ / | | / _ \\
  | |\  || __|| |_| | | |  / _ | >  <  | | | __|
  |_| \_|\___| \__,_||___| \_,_|/_/\_\ |__|\___|


 Learn more:
 - https://www.neuraxle.org/stable/index.html

 Contribute:
 - https://gitter.im/Neuraxle/community

 Open issue:
 - https://github.com/Neuraxio/Neuraxle

 Ask questions:
 - https://stackoverflow.com/questions/tagged/neuraxle
____________________________________________________________________
""")



"""
Pipeline Steps For Looping
=====================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""
import copy
from typing import List
from typing import Tuple

import numpy as np

from neuraxle.base import MetaStep, BaseStep, DataContainer, ExecutionContext, ResumableStepMixin, \
    ForceHandleOnlyMixin, ForceHandleMixin, TruncableJoblibStepSaver, NamedTupleList, BaseTransformer, MetaStepMixin
from neuraxle.data_container import ListDataContainer


class ForEachDataInput(ForceHandleOnlyMixin, ResumableStepMixin, MetaStep):
    """
    Truncable step that fits/transforms each step for each of the data inputs, and expected outputs.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.BaseSaver`,
        :class:`~neuraxle.base.BaseHasher`,
        :class:`~neuraxle.base.ResumableStepMixin`,
        :class:`~neuraxle.base.NonTransformableMixin`,
        :class:`~neuraxle.pipeline.Pipeline`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`~neuraxle.data_container.DataContainer`
    """

    def __init__(self, wrapped: BaseTransformer, cache_folder_when_no_handle=None):
        MetaStep.__init__(self, wrapped)
        ForceHandleOnlyMixin.__init__(self, cache_folder_when_no_handle)

    def _fit_data_container(self, data_container: DataContainer, context: ExecutionContext) -> BaseStep:
        """
        Fit each step for each data inputs, and expected outputs

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: self
        """
        for current_id, di, eo in data_container:
            self.wrapped = self.wrapped.handle_fit(
                DataContainer(data_inputs=di, current_ids=None, expected_outputs=eo),
                context
            )
        return self

    def _transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        Transform each step for each data inputs.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: self
        """
        output_data_container: ListDataContainer = ListDataContainer.empty(original_data_container=data_container)

        for current_id, di, eo in data_container:
            output: DataContainer = self.wrapped.handle_transform(
                DataContainer(data_inputs=di, current_ids=None, expected_outputs=eo),
                context
            )

            output_data_container.append(
                current_id,
                output.data_inputs,
                output.expected_outputs
            )
        output_data_container.summary_id = data_container.summary_id

        return output_data_container

    def _fit_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> Tuple[BaseStep, DataContainer]:
        """
        Fit transform each step for each data inputs, and expected outputs

        :param data_container: data container to fit transform
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext

        :return: self, transformed_data_container
        """
        output_data_container: DataContainer = ListDataContainer.empty(original_data_container=data_container)

        for current_id, di, eo in data_container:
            self.wrapped, output = self.wrapped.handle_fit_transform(
                DataContainer(data_inputs=di, current_ids=None, expected_outputs=eo),
                context
            )

            output_data_container.append(
                current_id,
                output.data_inputs,
                output.expected_outputs
            )

        output_data_container.summary_id = data_container.summary_id

        return self, output_data_container

    def hash_data_container(self, data_container: DataContainer) -> DataContainer:
        output_data_container = self.wrapped.hash_data_container(data_container)
        output_data_container.summary_id = data_container.summary_id

        return output_data_container

    def should_resume(self, data_container: DataContainer, context: ExecutionContext) -> bool:
        context: ExecutionContext = context.push(self)

        if isinstance(self.wrapped, ResumableStepMixin) and self.wrapped.should_resume(data_container, context):
            return True
        return False


class StepClonerForEachDataInput(ForceHandleOnlyMixin, MetaStep):
    def __init__(self, wrapped: BaseTransformer, copy_op=copy.deepcopy, cache_folder_when_no_handle=None):
        MetaStep.__init__(self, wrapped=wrapped)
        ForceHandleOnlyMixin.__init__(self, cache_folder_when_no_handle)
        self.savers.append(TruncableJoblibStepSaver())

        self.set_step(wrapped)
        self.steps_as_tuple: List[NamedTupleList] = []
        self.copy_op = copy_op

    def get_children(self) -> List[BaseStep]:
        """
        Get the list of all the children for that step.

        :return: list of children
        """
        children: List[BaseStep] = MetaStep.get_children(self)
        cloned_children = [step for _, step in self.steps_as_tuple]
        children.extend(cloned_children)
        return children

    def _will_process(self, data_container: DataContainer, context: ExecutionContext) -> ('BaseStep', DataContainer):
        data_container, context = super()._will_process(data_container, context)

        if len(self.steps_as_tuple) != len(data_container.data_inputs):
            self._copy_one_step_per_data_input(data_container)

        return data_container, context

    def _copy_one_step_per_data_input(self, data_container):
        # One copy of step per data input:
        steps = [self.copy_op(self.wrapped).set_name('{}[{}]'.format(self.wrapped.name, i)) for i in range(len(data_container))]
        self.steps_as_tuple = [(step.name, step) for step in steps]
        self._invalidate()

    def _fit_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> ('BaseStep', DataContainer):
        fitted_steps_data_containers = []
        for i, (current_ids, data_inputs, expected_outputs) in enumerate(data_container):
            fitted_step_data_container = self[i].handle_fit_transform(
                DataContainer(current_ids=current_ids, data_inputs=data_inputs, expected_outputs=expected_outputs),
                context
            )
            fitted_steps_data_containers.append(fitted_step_data_container)

        self.steps_as_tuple = [(step.name, step) for step, _ in fitted_steps_data_containers]

        output_data_container = ListDataContainer.empty()
        for _, data_container_batch in fitted_steps_data_containers:
            output_data_container.append_data_container(data_container_batch)

        return self, output_data_container

    def _fit_data_container(self, data_container: DataContainer, context: ExecutionContext) -> ('BaseStep', DataContainer):
        fitted_steps = []
        for i, (current_ids, data_inputs, expected_outputs) in enumerate(data_container):
            fitted_step = self[i].handle_fit(
                DataContainer(current_ids=current_ids, data_inputs=data_inputs, expected_outputs=expected_outputs),
                context
            )
            fitted_steps.append(fitted_step)

        self.steps_as_tuple = [(step.name, step) for step in fitted_steps]

        return self

    def _transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> ('BaseStep', DataContainer):
        transform_results = []
        for i, (current_ids, data_inputs, expected_outputs) in enumerate(data_container):
            transform_result = self[i].handle_transform(
                DataContainer(current_ids=current_ids, data_inputs=data_inputs, expected_outputs=expected_outputs),
                context
            )
            transform_results.append(transform_result)

        output_data_container = ListDataContainer.empty()
        for data_container_batch in transform_results:
            output_data_container.append_data_container(data_container_batch)
        return output_data_container

    def _inverse_transform_data_container(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        inverse_transform_results = []
        for i, (current_ids, data_inputs, expected_outputs) in enumerate(data_container):
            inverse_transform_result = self[i].handle_inverse_transform(
                DataContainer(current_ids=current_ids, data_inputs=data_inputs, expected_outputs=expected_outputs),
                context
            )
            inverse_transform_results.append(inverse_transform_result)

        output_data_container = ListDataContainer.empty()
        for data_container_batch in inverse_transform_results:
            output_data_container.append_data_container(data_container_batch)
        return output_data_container

    def inverse_transform(self, data_output):
        return [self[i].inverse_transform(di) for i, di in enumerate(data_output)]

    def __getitem__(self, item):
        """
        Get cloned step at the given index.

        :return: iter(self.steps_as_tuple)
        """
        return self.steps_as_tuple[item][1]

    def __iter__(self):
        """
        Iterate through the steps.

        :return: iter(self.steps_as_tuple)
        """
        return iter(self.steps_as_tuple)

    def __len__(self):
        """
        Get number of steps cloned for each data input.

        :return: len(self.steps_as_tuple)
        """
        return len(self.steps_as_tuple)


class FlattenForEach(ForceHandleMixin, ResumableStepMixin, MetaStep):
    """
    Step that reduces a dimension instead of manually looping on it.

    .. seealso::
        :class:`~neuraxle.base.BaseStep`,
        :class:`~neuraxle.base.BaseSaver`,
        :class:`~neuraxle.base.BaseHasher`,
        :class:`~neuraxle.base.ResumableStepMixin`,
        :class:`~neuraxle.base.MetaStepMixin`,
        :class:`~neuraxle.base.NonTransformableMixin`,
        :class:`~neuraxle.pipeline.Pipeline`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSamples`,
        :class:`~neuraxle.hyperparams.space.HyperparameterSpace`,
        :class:`~neuraxle.data_container.DataContainer`
    """

    def __init__(
            self,
            wrapped: BaseTransformer,
            then_unflatten: bool = True
    ):
        MetaStep.__init__(self, wrapped)
        ResumableStepMixin.__init__(self)
        ForceHandleMixin.__init__(self)

        self.then_unflatten = then_unflatten

        self.len_di = []
        self.len_eo = []

    def _will_process(self, data_container: DataContainer, context: ExecutionContext) -> (
    'BaseTransformer', DataContainer):
        """
        Flatten data container before any processing is done on the wrapped step.

        :param data_container: data container to flatten
        :param context: execution context
        :return: (data container, execution context)
        :rtype: ('BaseTransformer', DataContainer)
        """
        data_container, context = super()._will_process(data_container, context)

        if data_container.expected_outputs is None:
            expected_outputs = np.empty_like(np.array(data_container.data_inputs))
            expected_outputs.fill(np.nan)
            data_container.set_expected_outputs(expected_outputs)

        di, self.len_di = self._flatten_list(data_container.data_inputs)
        eo, self.len_eo = self._flatten_list(data_container.expected_outputs)

        flattened_data_container = DataContainer(
            summary_id=data_container.summary_id,
            data_inputs=di,
            expected_outputs=eo,
            sub_data_containers=data_container.sub_data_containers
        )

        return flattened_data_container, context

    def _flatten_list(self, list_to_flatten):
        """
        Flatten the first dimension of a list.

        :param list_to_flatten: list to flatten
        :return: flattened list, len flattened lists
        """
        if not isinstance(list_to_flatten, np.ndarray):
            list_to_flatten = np.array(list_to_flatten)

        if len(list_to_flatten.shape) == 1:
            return list_to_flatten, [1 for x in list_to_flatten]

        list_to_flatten = list(list_to_flatten)
        list_to_flatten = [list(x) for x in list_to_flatten]
        len_list_to_flatten = [len(x) for x in list_to_flatten]
        flattened_list = sum(list_to_flatten, [])

        return flattened_list, len_list_to_flatten

    def _did_process(self, data_container: DataContainer, context: ExecutionContext) -> DataContainer:
        """
        Reaugment the flattened data container.

        :param data_container: data container to then_unflatten
        :param context: execution context
        :return: data container
        """
        data_container = super()._did_process(data_container, context)

        if self.then_unflatten:
            data_container.set_data_inputs(self._reaugment_list(data_container.data_inputs, self.len_di))
            data_container.set_expected_outputs(self._reaugment_list(data_container.expected_outputs, self.len_eo))
            self.len_di = []
            self.len_eo = []

        return data_container

    def _reaugment_list(self, list_to_reaugment, flattened_dimension_lengths):
        """
        Reaugment list with the flattened dimension lengths.

        :param list_to_reaugment: list to then_unflatten
        :return: reaugmented numpy array
        """
        if not self.then_unflatten or list_to_reaugment is None:
            return list_to_reaugment

        reaugmented_list = []
        i = 0
        for list_length in flattened_dimension_lengths:
            sub_list = list_to_reaugment[i:i + list_length]
            reaugmented_list.append(sub_list)
            i += list_length

        return reaugmented_list

    def should_resume(self, data_container: DataContainer, context: ExecutionContext):
        context = context.push(self)

        if isinstance(self.wrapped, ResumableStepMixin) and self.wrapped.should_resume(data_container, context):
            return True
        return False



"""
Neuraxle's API classes
========================================
Neuraxle's high-level API classes. Useful to make complex Deep Learning pipelines by calling just a few minimal things.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
from typing import Dict, Callable, Union

from neuraxle.base import BaseStep, NamedTupleList, EvaluableStepMixin, ForceHandleMixin, MetaStep
from neuraxle.metaopt.random import ValidationSplitWrapper
from neuraxle.metrics import MetricsWrapper
from neuraxle.pipeline import MiniBatchSequentialPipeline
from neuraxle.steps.data import EpochRepeater, TrainShuffled

VALIDATION_SPLIT_STEP_NAME = 'validation_split_wrapper'
EPOCH_METRICS_STEP_NAME = 'epoch_metrics'
BATCH_METRICS_STEP_NAME = 'batch_metrics'


class DeepLearningPipeline(EvaluableStepMixin, ForceHandleMixin, MetaStep):
    """
    Adds an epoch loop, a validation split, and mini batching to a pipeline.
    It also tracks batch metrics, and epoch metrics.


    Example usage :

    .. code-block:: python

        p = DeepLearningPipeline(
            pipeline,
            validation_size=VALIDATION_SIZE,
            batch_size=BATCH_SIZE,
            batch_metrics={'mse': to_numpy_metric_wrapper(mean_squared_error)},
            shuffle_in_each_epoch_at_train=True,
            n_epochs=N_EPOCHS,
            epochs_metrics={'mse': to_numpy_metric_wrapper(mean_squared_error)},
            scoring_function=to_numpy_metric_wrapper(mean_squared_error),
        )

        p, outputs = p.fit_transform(data_inputs, expected_outputs)

        batch_mse_train = p.get_batch_metric_train('mse')
        epoch_mse_train = p.get_epoch_metric_train('mse')
        batch_mse_validation = p.get_batch_metric_validation('mse')
        epoch_mse_validation = p.get_epoch_metric_validation('mse')

    It uses :class:`EpochRepeater`, :class:`ValidationSplitWrapper`, and :class:`MiniBatchSequentialPipeline`

    .. seealso::
        :class:`~neuraxle.steps.data.EpochRepeater`,
        :class:`~neuraxle.metaopt.random.ValidationSplitWrapper`,
        :class:`~neuraxle.pipeline.MiniBatchSequentialPipeline`,
        :class:`~neuraxle.pipeline.Pipeline`,
        :class:`~neuraxle.pipeline.CustomPipelineMixin`,
        :class:`~neuraxle.metrics.MetricsWrapper`
    """

    def __init__(
            self,
            pipeline: Union[BaseStep, NamedTupleList],
            validation_size: float = None,
            batch_size: int = None,
            batch_metrics: Dict[str, Callable] = None,
            shuffle_in_each_epoch_at_train: bool = True,
            seed: int = None,
            n_epochs: int = 1,
            epochs_metrics: Dict[str, Callable] = None,
            scoring_function: Callable = None,
            cache_folder: str = None,
            print_epoch_metrics=False,
            print_batch_metrics=False
    ):
        """
        :param pipeline: pipeline to wrap with an epoch repeater, a validation split wrapper, and a mini batch sequential pipeline
        :param validation_size: ratio for validation size between 0 and 1
        :param batch_size: batch size for the mini batch sequential pipeline
        :param batch_metrics: metrics to calculate for each processed mini batch
        :param shuffle_in_each_epoch_at_train:
        :param seed: random seed for the data shuffling that can be done at each epoch when the param shuffle_in_each_epoch_at_train is True
        :param n_epochs: number of epochs
        :param epochs_metrics: metrics to calculate for each epoch
        :param scoring_function: scoring function with two arguments (y_true, y_pred)
        :param cache_folder: cache folder to be used inside the pipeline
        :param print_epoch_metrics: whether or not to print epoch metrics
        :param print_batch_metrics: whether or not to print batch metrics
        """
        if epochs_metrics is None:
            epochs_metrics = {}
        if batch_metrics is None:
            batch_metrics = {}

        self.final_scoring_metric = scoring_function
        self.epochs_metrics = epochs_metrics
        self.n_epochs = n_epochs
        self.shuffle_in_each_epoch_at_train = shuffle_in_each_epoch_at_train
        self.batch_size = batch_size
        self.batch_metrics = batch_metrics
        self.validation_size = validation_size
        self.print_batch_metrics = print_batch_metrics
        self.print_epoch_metrics = print_epoch_metrics

        wrapped: BaseStep = pipeline
        wrapped: MiniBatchSequentialPipeline = self._create_mini_batch_pipeline(wrapped)

        if shuffle_in_each_epoch_at_train:
            wrapped = TrainShuffled(wrapped=wrapped, seed=seed)

        wrapped = self._create_validation_split(wrapped)
        wrapped = self._create_epoch_repeater(wrapped)

        MetaStep.__init__(self, wrapped)
        EvaluableStepMixin.__init__(self)
        ForceHandleMixin.__init__(self, cache_folder)

    def _create_mini_batch_pipeline(self, wrapped: BaseStep) -> BaseStep:
        """
        Add mini batching and batch metrics by wrapping the step with :class:`MetricsWrapper`, and  :class:̀MiniBatchSequentialPipeline`.

        :param wrapped: pipeline step
        :type wrapped: BaseStep
        :return: wrapped pipeline step
        :rtype: MetricsWrapper
        """
        if self.batch_size is not None:
            wrapped = MetricsWrapper(wrapped=wrapped, metrics=self.batch_metrics, name=BATCH_METRICS_STEP_NAME,
                                     print_metrics=self.print_batch_metrics)
            wrapped = MiniBatchSequentialPipeline(
                [wrapped],
                batch_size=self.batch_size
            )

        return wrapped

    def _create_validation_split(self, wrapped: BaseStep) -> BaseStep:
        """
        Add validation split and epoch metrics by wrapping the step with :class:`MetricsWrapper`, and  :class:̀ValidationSplitWrapper`.

        :param wrapped: pipeline step
        :type wrapped: BaseStep
        :return: wrapped pipeline step
        :rtype: MetricsWrapper
        """
        wrapped = MetricsWrapper(
            wrapped=wrapped,
            metrics=self.epochs_metrics,
            name=EPOCH_METRICS_STEP_NAME,
            print_metrics=self.print_epoch_metrics
        )

        if self.validation_size is not None:
            wrapped = ValidationSplitWrapper(
                wrapped=wrapped,
                test_size=self.validation_size,
                scoring_function=self.final_scoring_metric
            ).set_name(VALIDATION_SPLIT_STEP_NAME)

        return wrapped

    def _create_epoch_repeater(self, wrapped: BaseStep) -> BaseStep:
        """
        Add epoch loop by wrapping the step with :class:`EpochRepeater`.

        :param wrapped: pipeline step
        :type wrapped: BaseStep
        :return: wrapped pipeline step
        :rtype: BaseStep
        """
        if self.n_epochs is not None:
            wrapped = EpochRepeater(wrapped, epochs=self.n_epochs, repeat_in_test_mode=False)
        return wrapped

    def get_score(self):
        """
        Get latest score. This function had to be defined for the hyperparameter optimization steps.

        :return: score
        :rtype: float
        """
        return self.get_step_by_name(VALIDATION_SPLIT_STEP_NAME).get_score()

    def get_score_validation(self):
        """
        Get latest score validation.

        :return: score
        :rtype: float
        """
        return self.get_step_by_name(VALIDATION_SPLIT_STEP_NAME).get_score_validation()

    def get_score_train(self) -> float:
        """
        Get latest score train.

        :return: score
        :rtype: float
        """
        return self.get_step_by_name(VALIDATION_SPLIT_STEP_NAME).get_score_train()

